# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

r"""Script utils for generating datasets figures and dataframes."""

import concurrent.futures
import functools
import itertools
import multiprocessing
import os
import traceback
from typing import Any, Callable, List, Optional, TypeVar

from absl import app
from absl import logging
import tensorflow_datasets as tfds
from tensorflow_datasets.core.utils.lazy_imports_utils import tensorflow as tf

# pylint: disable=logging-format-interpolation,logging-not-lazy,logging-fstring-interpolation

T = TypeVar('T')

_WORKER_COUNT_DATASETS = 10


def _log_exception(fn):
  """Logs the exceptions from a `ThreadPoolExecutor`."""

  @functools.wraps(fn)
  def decorated(*args, **kwargs):
    try:
      return fn(*args, **kwargs)
    except Exception:  # pylint: disable=broad-except
      err_str = traceback.format_exc()
      logging.error(f'Exception occurred for {args}, {kwargs}:\n' + err_str)
      raise

  return decorated


@_log_exception
def generate_and_save_artifact(
    full_name: str,
    *,
    dst_dir: tfds.typing.PathLike,
    overwrite: bool,
    file_extension: str,
    get_artifact_fn: Callable[[tf.data.Dataset, tfds.core.DatasetInfo], T],
    save_artifact_fn: Callable[[str, T], None],
) -> None:
  """Builds and saves the generated artifact for the dataset in dst_dir.

  Args:
    full_name: Name of the dataset to build `dataset`, `dataset/config`.
    dst_dir: Destination where the dataset will be saved (as
      `dataset-config-version`)
    overwrite: If True, recompute the image even if it exists.
    file_extension: File extension of the artifact (e.g. `.png`)
    get_artifact_fn: Function which extracts the dataset artifact to save.
    save_artifact_fn: Function which save the extracted artifact.
  """
  dst_filename = full_name.replace('/', '-') + file_extension
  dst_path = os.path.join(dst_dir, dst_filename)
  # If the file already exists, skip the generation
  if not overwrite and tf.io.gfile.exists(dst_path):
    logging.info(f'Skipping generation for {full_name} (already exists)')
    return

  logging.info(f'Generating for {full_name}...')
  # Load the dataset.
  builder_name, _, version = full_name.rpartition('/')
  builder = tfds.builder(f'{builder_name}:{version}')
  split_names = list(builder.info.splits.keys())
  if not split_names:
    logging.info(f'Dataset `{full_name}` not generated.')
    return
  elif 'train' in split_names:
    split = 'train'
  else:
    split = split_names[0]
  ds = builder.as_dataset(split=split, shuffle_files=False)
  if not tf.io.gfile.exists(dst_dir):
    tf.io.gfile.makedirs(dst_dir)
  try:
    artifact = get_artifact_fn(ds.take(10), builder.info)
  except Exception:  # pylint: disable=broad-except
    logging.info(f'Generation not supported for dataset `{full_name}`')
    return

  save_artifact_fn(dst_path, artifact)


def _get_full_names(datasets: Optional[List[str]] = None) -> List[str]:
  """Lists all builder names `ds/version` and `ds/config/version` to generate.

  Args:
    datasets: List of datasets from which get the builder names.

  Returns:
    builder_names: The builder names.
  """
  if datasets is None:
    return tfds.core.load.list_full_names(current_version_only=True)
  else:
    builder_names = list(
        itertools.chain.from_iterable(
            [
                tfds.core.load.single_full_names(builder_name)
                for builder_name in datasets
            ]
        )
    )
    return builder_names


def multi_process_map(
    worker_fn: Callable[..., None],
    datasets: Optional[List[str]] = None,
) -> None:
  """Applies the function for each given datasets.

  Args:
    worker_fn: Function called on each dataset version.
    datasets: List of all `dataset` names to generate. If None, visualization
      for all available datasets will be generated.
  """
  full_names = _get_full_names(datasets)
  logging.info(f'Generate figures for {len(full_names)} builders')
  with multiprocessing.Pool(_WORKER_COUNT_DATASETS) as tpool:
    list(tpool.map(worker_fn, full_names))


def multi_thread_map(
    worker_fn: Callable[..., None],
    datasets: Optional[List[str]] = None,
) -> None:
  """Applies the function for each given datasets.

  Args:
    worker_fn: Function called on each dataset version.
    datasets: List of all `dataset` names to generate. If None, visualization
      for all available datasets will be generated.
  """
  full_names = _get_full_names(datasets)
  with concurrent.futures.ThreadPoolExecutor(
      max_workers=_WORKER_COUNT_DATASETS,
  ) as executor:
    list(executor.map(worker_fn, full_names))


def multi_process_run(main: Callable[[Any], None]) -> None:
  """Same as `absl.app.run` but with special multiprocess flags."""
  app.run(main)



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for WIDER FACE dataset module."""

from tensorflow_datasets import testing
from tensorflow_datasets.object_detection import wider_face
import tensorflow_datasets.public_api as tfds


class WiderFaceTest(testing.DatasetBuilderTestCase):
  DATASET_CLASS = wider_face.WiderFace
  SPLITS = {
      tfds.Split.TRAIN: 3,
      tfds.Split.VALIDATION: 3,
      tfds.Split.TEST: 3,
  }
  DL_EXTRACT_RESULT = {
      'wider_train': 'wider_train',
      'wider_val': 'wider_val',
      'wider_test': 'wider_test',
      'wider_annot': 'wider_annot',
  }


if __name__ == '__main__':
  testing.test_main()



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Largest Covid-19 epidemiological database with a powerful set of expansive covariates.

It includes open, publicly sourced, licensed data relating to demographics,
economy, epidemiology, geography, health, hospitalizations, mobility, government
response, weather, and more.
"""

import numpy as np
from tensorflow_datasets.core.utils.lazy_imports_utils import tensorflow as tf
import tensorflow_datasets.public_api as tfds

_DESCRIPTION = """
This repository attempts to assemble the largest Covid-19 epidemiological database
in addition to a powerful set of expansive covariates. It includes open, publicly sourced,
licensed data relating to demographics, economy, epidemiology, geography, health, hospitalizations,
mobility, government response, weather, and more.

This particular dataset corresponds to a join of all the different tables that are part of the repository.
Therefore, expect the resulting samples to be highly sparse.
"""

_CITATION = """
@article{Wahltinez2020,
  author = "O. Wahltinez and others",
  year = 2020,
  title = "COVID-19 Open-Data: curating a fine-grained, global-scale data repository for SARS-CoV-2",
  note = "Work in progress",
  url = {https://goo.gle/covid-19-open-data},
}
"""

_N_RECORDS = 11266423  # It should be an upper bound
_BATCH_SIZE = 10000


class Covid19(tfds.core.GeneratorBasedBuilder):
  """DatasetBuilder for covid19 dataset."""

  VERSION = tfds.core.Version('1.0.0')
  RELEASE_NOTES = {
      '1.0.0': 'Initial release.',
  }

  def _info(self) -> tfds.core.DatasetInfo:
    """Returns the dataset metadata."""
    return tfds.core.DatasetInfo(
        builder=self,
        description=_DESCRIPTION,
        features=tfds.features.FeaturesDict({
            'location_key': np.str_,
            'date': np.str_,
            'place_id': np.str_,
            'wikidata_id': np.str_,
            'datacommons_id': np.str_,
            'country_code': np.str_,
            'country_name': np.str_,
            'iso_3166_1_alpha_2': np.str_,
            'iso_3166_1_alpha_3': np.str_,
            'aggregation_level': np.float64,
            'new_confirmed': np.float64,
            'new_deceased': np.float64,
            'cumulative_confirmed': np.float64,
            'cumulative_deceased': np.float64,
            'cumulative_tested': np.float64,
            'new_persons_vaccinated': np.float64,
            'cumulative_persons_vaccinated': np.float64,
            'new_persons_fully_vaccinated': np.float64,
            'cumulative_persons_fully_vaccinated': np.float64,
            'new_vaccine_doses_administered': np.float64,
            'cumulative_vaccine_doses_administered': np.float64,
            'population': np.float64,
            'population_male': np.float64,
            'population_female': np.float64,
            'population_rural': np.float64,
            'population_urban': np.float64,
            'population_density': np.float64,
            'human_development_index': np.float64,
            'population_age_00_09': np.float64,
            'population_age_10_19': np.float64,
            'population_age_20_29': np.float64,
            'population_age_30_39': np.float64,
            'population_age_40_49': np.float64,
            'population_age_50_59': np.float64,
            'population_age_60_69': np.float64,
            'population_age_70_79': np.float64,
            'population_age_80_and_older': np.float64,
            'gdp_usd': np.float64,
            'gdp_per_capita_usd': np.float64,
            'openstreetmap_id': np.str_,
            'latitude': np.float64,
            'longitude': np.float64,
            'area_sq_km': np.float64,
            'smoking_prevalence': np.float64,
            'diabetes_prevalence': np.float64,
            'infant_mortality_rate': np.float64,
            'nurses_per_1000': np.float64,
            'physicians_per_1000': np.float64,
            'health_expenditure_usd': np.float64,
            'out_of_pocket_health_expenditure_usd': np.float64,
            'school_closing': np.float64,
            'workplace_closing': np.float64,
            'cancel_public_events': np.float64,
            'restrictions_on_gatherings': np.float64,
            'public_transport_closing': np.float64,
            'stay_at_home_requirements': np.float64,
            'restrictions_on_internal_movement': np.float64,
            'international_travel_controls': np.float64,
            'income_support': np.float64,
            'debt_relief': np.float64,
            'fiscal_measures': np.float64,
            'international_support': np.float64,
            'public_information_campaigns': np.float64,
            'testing_policy': np.float64,
            'contact_tracing': np.float64,
            'emergency_investment_in_healthcare': np.float64,
            'investment_in_vaccines': np.float64,
            'facial_coverings': np.float64,
            'vaccination_policy': np.float64,
            'stringency_index': np.float64,
            'average_temperature_celsius': np.float64,
            'minimum_temperature_celsius': np.float64,
            'maximum_temperature_celsius': np.float64,
            'rainfall_mm': np.float64,
            'snowfall_mm': np.float64,
            'dew_point': np.float64,
            'relative_humidity': np.float64,
            'new_tested': np.float64,
            'population_largest_city': np.float64,
            'population_clustered': np.float64,
            'human_capital_index': np.float64,
            'area_rural_sq_km': np.float64,
            'area_urban_sq_km': np.float64,
            'life_expectancy': np.float64,
            'adult_male_mortality_rate': np.float64,
            'adult_female_mortality_rate': np.float64,
            'pollution_mortality_rate': np.float64,
            'comorbidity_mortality_rate': np.float64,
            'mobility_retail_and_recreation': np.float64,
            'mobility_grocery_and_pharmacy': np.float64,
            'mobility_parks': np.float64,
            'mobility_transit_stations': np.float64,
            'mobility_workplaces': np.float64,
            'mobility_residential': np.float64,
            'hospital_beds_per_1000': np.float64,
            'subregion1_code': np.str_,
            'subregion1_name': np.str_,
            'new_recovered': np.float64,
            'cumulative_recovered': np.float64,
            'elevation_m': np.float64,
            'new_hospitalized_patients': np.float64,
            'cumulative_hospitalized_patients': np.float64,
            'new_intensive_care_patients': np.float64,
            'cumulative_intensive_care_patients': np.float64,
            'new_confirmed_age_0': np.float64,
            'new_confirmed_age_1': np.float64,
            'new_confirmed_age_2': np.float64,
            'new_confirmed_age_3': np.float64,
            'new_confirmed_age_4': np.float64,
            'new_confirmed_age_5': np.float64,
            'new_confirmed_age_6': np.float64,
            'new_confirmed_age_7': np.float64,
            'new_confirmed_age_8': np.float64,
            'new_confirmed_age_9': np.float64,
            'cumulative_confirmed_age_0': np.float64,
            'cumulative_confirmed_age_1': np.float64,
            'cumulative_confirmed_age_2': np.float64,
            'cumulative_confirmed_age_3': np.float64,
            'cumulative_confirmed_age_4': np.float64,
            'cumulative_confirmed_age_5': np.float64,
            'cumulative_confirmed_age_6': np.float64,
            'cumulative_confirmed_age_7': np.float64,
            'cumulative_confirmed_age_8': np.float64,
            'cumulative_confirmed_age_9': np.float64,
            'new_deceased_age_0': np.float64,
            'new_deceased_age_1': np.float64,
            'new_deceased_age_2': np.float64,
            'new_deceased_age_3': np.float64,
            'new_deceased_age_4': np.float64,
            'new_deceased_age_5': np.float64,
            'new_deceased_age_6': np.float64,
            'new_deceased_age_7': np.float64,
            'new_deceased_age_8': np.float64,
            'new_deceased_age_9': np.float64,
            'cumulative_deceased_age_0': np.float64,
            'cumulative_deceased_age_1': np.float64,
            'cumulative_deceased_age_2': np.float64,
            'cumulative_deceased_age_3': np.float64,
            'cumulative_deceased_age_4': np.float64,
            'cumulative_deceased_age_5': np.float64,
            'cumulative_deceased_age_6': np.float64,
            'cumulative_deceased_age_7': np.float64,
            'cumulative_deceased_age_8': np.float64,
            'cumulative_deceased_age_9': np.float64,
            'new_tested_age_0': np.float64,
            'new_tested_age_1': np.float64,
            'new_tested_age_2': np.float64,
            'new_tested_age_3': np.float64,
            'new_tested_age_4': np.float64,
            'new_tested_age_5': np.float64,
            'new_tested_age_6': np.float64,
            'new_tested_age_7': np.float64,
            'new_tested_age_8': np.float64,
            'new_tested_age_9': np.float64,
            'cumulative_tested_age_0': np.float64,
            'cumulative_tested_age_1': np.float64,
            'cumulative_tested_age_2': np.float64,
            'cumulative_tested_age_3': np.float64,
            'cumulative_tested_age_4': np.float64,
            'cumulative_tested_age_5': np.float64,
            'cumulative_tested_age_6': np.float64,
            'cumulative_tested_age_7': np.float64,
            'cumulative_tested_age_8': np.float64,
            'cumulative_tested_age_9': np.float64,
            'new_hospitalized_patients_age_0': np.float64,
            'new_hospitalized_patients_age_1': np.float64,
            'new_hospitalized_patients_age_2': np.float64,
            'new_hospitalized_patients_age_3': np.float64,
            'new_hospitalized_patients_age_4': np.float64,
            'new_hospitalized_patients_age_5': np.float64,
            'new_hospitalized_patients_age_6': np.float64,
            'new_hospitalized_patients_age_7': np.float64,
            'new_hospitalized_patients_age_8': np.float64,
            'new_hospitalized_patients_age_9': np.float64,
            'cumulative_hospitalized_patients_age_0': np.float64,
            'cumulative_hospitalized_patients_age_1': np.float64,
            'cumulative_hospitalized_patients_age_2': np.float64,
            'cumulative_hospitalized_patients_age_3': np.float64,
            'cumulative_hospitalized_patients_age_4': np.float64,
            'cumulative_hospitalized_patients_age_5': np.float64,
            'cumulative_hospitalized_patients_age_6': np.float64,
            'cumulative_hospitalized_patients_age_7': np.float64,
            'cumulative_hospitalized_patients_age_8': np.float64,
            'cumulative_hospitalized_patients_age_9': np.float64,
            'new_intensive_care_patients_age_0': np.float64,
            'new_intensive_care_patients_age_1': np.float64,
            'new_intensive_care_patients_age_2': np.float64,
            'new_intensive_care_patients_age_3': np.float64,
            'new_intensive_care_patients_age_4': np.float64,
            'new_intensive_care_patients_age_5': np.float64,
            'new_intensive_care_patients_age_6': np.float64,
            'new_intensive_care_patients_age_7': np.float64,
            'new_intensive_care_patients_age_8': np.float64,
            'new_intensive_care_patients_age_9': np.float64,
            'cumulative_intensive_care_patients_age_0': np.float64,
            'cumulative_intensive_care_patients_age_1': np.float64,
            'cumulative_intensive_care_patients_age_2': np.float64,
            'cumulative_intensive_care_patients_age_3': np.float64,
            'cumulative_intensive_care_patients_age_4': np.float64,
            'cumulative_intensive_care_patients_age_5': np.float64,
            'cumulative_intensive_care_patients_age_6': np.float64,
            'cumulative_intensive_care_patients_age_7': np.float64,
            'cumulative_intensive_care_patients_age_8': np.float64,
            'cumulative_intensive_care_patients_age_9': np.float64,
            # Stored as string because values look like "30-39"
            'age_bin_0': np.str_,
            'age_bin_1': np.str_,
            'age_bin_2': np.str_,
            'age_bin_3': np.str_,
            'age_bin_4': np.str_,
            'age_bin_5': np.str_,
            'age_bin_6': np.str_,
            'age_bin_7': np.str_,
            'age_bin_8': np.str_,
            'age_bin_9': np.str_,
            'new_confirmed_male': np.float64,
            'new_confirmed_female': np.float64,
            'cumulative_confirmed_male': np.float64,
            'cumulative_confirmed_female': np.float64,
            'new_deceased_male': np.float64,
            'new_deceased_female': np.float64,
            'cumulative_deceased_male': np.float64,
            'cumulative_deceased_female': np.float64,
            'new_tested_male': np.float64,
            'new_tested_female': np.float64,
            'cumulative_tested_male': np.float64,
            'cumulative_tested_female': np.float64,
            'new_hospitalized_patients_male': np.float64,
            'new_hospitalized_patients_female': np.float64,
            'cumulative_hospitalized_patients_male': np.float64,
            'cumulative_hospitalized_patients_female': np.float64,
            'new_intensive_care_patients_male': np.float64,
            'new_intensive_care_patients_female': np.float64,
            'cumulative_intensive_care_patients_male': np.float64,
            'cumulative_intensive_care_patients_female': np.float64,
            'subregion2_code': np.str_,
            'subregion2_name': np.str_,
            'current_hospitalized_patients': np.float64,
            'current_intensive_care_patients': np.float64,
            'current_ventilator_patients': np.float64,
            'search_trends_abdominal_obesity': np.float64,
            'search_trends_abdominal_pain': np.float64,
            'search_trends_acne': np.float64,
            'search_trends_actinic_keratosis': np.float64,
            'search_trends_acute_bronchitis': np.float64,
            'search_trends_adrenal_crisis': np.float64,
            'search_trends_ageusia': np.float64,
            'search_trends_alcoholism': np.float64,
            'search_trends_allergic_conjunctivitis': np.float64,
            'search_trends_allergy': np.float64,
            'search_trends_amblyopia': np.float64,
            'search_trends_amenorrhea': np.float64,
            'search_trends_amnesia': np.float64,
            'search_trends_anal_fissure': np.float64,
            'search_trends_anaphylaxis': np.float64,
            'search_trends_anemia': np.float64,
            'search_trends_angina_pectoris': np.float64,
            'search_trends_angioedema': np.float64,
            'search_trends_angular_cheilitis': np.float64,
            'search_trends_anosmia': np.float64,
            'search_trends_anxiety': np.float64,
            'search_trends_aphasia': np.float64,
            'search_trends_aphonia': np.float64,
            'search_trends_apnea': np.float64,
            'search_trends_arthralgia': np.float64,
            'search_trends_arthritis': np.float64,
            'search_trends_ascites': np.float64,
            'search_trends_asperger_syndrome': np.float64,
            'search_trends_asphyxia': np.float64,
            'search_trends_asthma': np.float64,
            'search_trends_astigmatism': np.float64,
            'search_trends_ataxia': np.float64,
            'search_trends_atheroma': np.float64,
            'search_trends_attention_deficit_hyperactivity_disorder': (
                np.float64
            ),
            'search_trends_auditory_hallucination': np.float64,
            'search_trends_autoimmune_disease': np.float64,
            'search_trends_avoidant_personality_disorder': np.float64,
            'search_trends_back_pain': np.float64,
            'search_trends_bacterial_vaginosis': np.float64,
            'search_trends_balance_disorder': np.float64,
            'search_trends_beaus_lines': np.float64,
            'search_trends_bells_palsy': np.float64,
            'search_trends_biliary_colic': np.float64,
            'search_trends_binge_eating': np.float64,
            'search_trends_bleeding': np.float64,
            'search_trends_bleeding_on_probing': np.float64,
            'search_trends_blepharospasm': np.float64,
            'search_trends_bloating': np.float64,
            'search_trends_blood_in_stool': np.float64,
            'search_trends_blurred_vision': np.float64,
            'search_trends_blushing': np.float64,
            'search_trends_boil': np.float64,
            'search_trends_bone_fracture': np.float64,
            'search_trends_bone_tumor': np.float64,
            'search_trends_bowel_obstruction': np.float64,
            'search_trends_bradycardia': np.float64,
            'search_trends_braxton_hicks_contractions': np.float64,
            'search_trends_breakthrough_bleeding': np.float64,
            'search_trends_breast_pain': np.float64,
            'search_trends_bronchitis': np.float64,
            'search_trends_bruise': np.float64,
            'search_trends_bruxism': np.float64,
            'search_trends_bunion': np.float64,
            'search_trends_burn': np.float64,
            'search_trends_burning_chest_pain': np.float64,
            'search_trends_burning_mouth_syndrome': np.float64,
            'search_trends_candidiasis': np.float64,
            'search_trends_canker_sore': np.float64,
            'search_trends_cardiac_arrest': np.float64,
            'search_trends_carpal_tunnel_syndrome': np.float64,
            'search_trends_cataplexy': np.float64,
            'search_trends_cataract': np.float64,
            'search_trends_chancre': np.float64,
            'search_trends_cheilitis': np.float64,
            'search_trends_chest_pain': np.float64,
            'search_trends_chills': np.float64,
            'search_trends_chorea': np.float64,
            'search_trends_chronic_pain': np.float64,
            'search_trends_cirrhosis': np.float64,
            'search_trends_cleft_lip_and_cleft_palate': np.float64,
            'search_trends_clouding_of_consciousness': np.float64,
            'search_trends_cluster_headache': np.float64,
            'search_trends_colitis': np.float64,
            'search_trends_coma': np.float64,
            'search_trends_common_cold': np.float64,
            'search_trends_compulsive_behavior': np.float64,
            'search_trends_compulsive_hoarding': np.float64,
            'search_trends_confusion': np.float64,
            'search_trends_congenital_heart_defect': np.float64,
            'search_trends_conjunctivitis': np.float64,
            'search_trends_constipation': np.float64,
            'search_trends_convulsion': np.float64,
            'search_trends_cough': np.float64,
            'search_trends_crackles': np.float64,
            'search_trends_cramp': np.float64,
            'search_trends_crepitus': np.float64,
            'search_trends_croup': np.float64,
            'search_trends_cyanosis': np.float64,
            'search_trends_dandruff': np.float64,
            'search_trends_delayed_onset_muscle_soreness': np.float64,
            'search_trends_dementia': np.float64,
            'search_trends_dentin_hypersensitivity': np.float64,
            'search_trends_depersonalization': np.float64,
            'search_trends_depression': np.float64,
            'search_trends_dermatitis': np.float64,
            'search_trends_desquamation': np.float64,
            'search_trends_developmental_disability': np.float64,
            'search_trends_diabetes': np.float64,
            'search_trends_diabetic_ketoacidosis': np.float64,
            'search_trends_diarrhea': np.float64,
            'search_trends_dizziness': np.float64,
            'search_trends_dry_eye_syndrome': np.float64,
            'search_trends_dysautonomia': np.float64,
            'search_trends_dysgeusia': np.float64,
            'search_trends_dysmenorrhea': np.float64,
            'search_trends_dyspareunia': np.float64,
            'search_trends_dysphagia': np.float64,
            'search_trends_dysphoria': np.float64,
            'search_trends_dystonia': np.float64,
            'search_trends_dysuria': np.float64,
            'search_trends_ear_pain': np.float64,
            'search_trends_eczema': np.float64,
            'search_trends_edema': np.float64,
            'search_trends_encephalitis': np.float64,
            'search_trends_encephalopathy': np.float64,
            'search_trends_epidermoid_cyst': np.float64,
            'search_trends_epilepsy': np.float64,
            'search_trends_epiphora': np.float64,
            'search_trends_erectile_dysfunction': np.float64,
            'search_trends_erythema': np.float64,
            'search_trends_erythema_chronicum_migrans': np.float64,
            'search_trends_esophagitis': np.float64,
            'search_trends_excessive_daytime_sleepiness': np.float64,
            'search_trends_eye_pain': np.float64,
            'search_trends_eye_strain': np.float64,
            'search_trends_facial_nerve_paralysis': np.float64,
            'search_trends_facial_swelling': np.float64,
            'search_trends_fasciculation': np.float64,
            'search_trends_fatigue': np.float64,
            'search_trends_fatty_liver_disease': np.float64,
            'search_trends_fecal_incontinence': np.float64,
            'search_trends_fever': np.float64,
            'search_trends_fibrillation': np.float64,
            'search_trends_fibrocystic_breast_changes': np.float64,
            'search_trends_fibromyalgia': np.float64,
            'search_trends_flatulence': np.float64,
            'search_trends_floater': np.float64,
            'search_trends_focal_seizure': np.float64,
            'search_trends_folate_deficiency': np.float64,
            'search_trends_food_craving': np.float64,
            'search_trends_food_intolerance': np.float64,
            'search_trends_frequent_urination': np.float64,
            'search_trends_gastroesophageal_reflux_disease': np.float64,
            'search_trends_gastroparesis': np.float64,
            'search_trends_generalized_anxiety_disorder': np.float64,
            'search_trends_genital_wart': np.float64,
            'search_trends_gingival_recession': np.float64,
            'search_trends_gingivitis': np.float64,
            'search_trends_globus_pharyngis': np.float64,
            'search_trends_goitre': np.float64,
            'search_trends_gout': np.float64,
            'search_trends_grandiosity': np.float64,
            'search_trends_granuloma': np.float64,
            'search_trends_guilt': np.float64,
            'search_trends_hair_loss': np.float64,
            'search_trends_halitosis': np.float64,
            'search_trends_hay_fever': np.float64,
            'search_trends_headache': np.float64,
            'search_trends_heart_arrhythmia': np.float64,
            'search_trends_heart_murmur': np.float64,
            'search_trends_heartburn': np.float64,
            'search_trends_hematochezia': np.float64,
            'search_trends_hematoma': np.float64,
            'search_trends_hematuria': np.float64,
            'search_trends_hemolysis': np.float64,
            'search_trends_hemoptysis': np.float64,
            'search_trends_hemorrhoids': np.float64,
            'search_trends_hepatic_encephalopathy': np.float64,
            'search_trends_hepatitis': np.float64,
            'search_trends_hepatotoxicity': np.float64,
            'search_trends_hiccup': np.float64,
            'search_trends_hip_pain': np.float64,
            'search_trends_hives': np.float64,
            'search_trends_hot_flash': np.float64,
            'search_trends_hydrocephalus': np.float64,
            'search_trends_hypercalcaemia': np.float64,
            'search_trends_hypercapnia': np.float64,
            'search_trends_hypercholesterolemia': np.float64,
            'search_trends_hyperemesis_gravidarum': np.float64,
            'search_trends_hyperglycemia': np.float64,
            'search_trends_hyperhidrosis': np.float64,
            'search_trends_hyperkalemia': np.float64,
            'search_trends_hyperlipidemia': np.float64,
            'search_trends_hypermobility': np.float64,
            'search_trends_hyperpigmentation': np.float64,
            'search_trends_hypersomnia': np.float64,
            'search_trends_hypertension': np.float64,
            'search_trends_hyperthermia': np.float64,
            'search_trends_hyperthyroidism': np.float64,
            'search_trends_hypertriglyceridemia': np.float64,
            'search_trends_hypertrophy': np.float64,
            'search_trends_hyperventilation': np.float64,
            'search_trends_hypocalcaemia': np.float64,
            'search_trends_hypochondriasis': np.float64,
            'search_trends_hypoglycemia': np.float64,
            'search_trends_hypogonadism': np.float64,
            'search_trends_hypokalemia': np.float64,
            'search_trends_hypomania': np.float64,
            'search_trends_hyponatremia': np.float64,
            'search_trends_hypotension': np.float64,
            'search_trends_hypothyroidism': np.float64,
            'search_trends_hypoxemia': np.float64,
            'search_trends_hypoxia': np.float64,
            'search_trends_impetigo': np.float64,
            'search_trends_implantation_bleeding': np.float64,
            'search_trends_impulsivity': np.float64,
            'search_trends_indigestion': np.float64,
            'search_trends_infection': np.float64,
            'search_trends_inflammation': np.float64,
            'search_trends_inflammatory_bowel_disease': np.float64,
            'search_trends_ingrown_hair': np.float64,
            'search_trends_insomnia': np.float64,
            'search_trends_insulin_resistance': np.float64,
            'search_trends_intermenstrual_bleeding': np.float64,
            'search_trends_intracranial_pressure': np.float64,
            'search_trends_iron_deficiency': np.float64,
            'search_trends_irregular_menstruation': np.float64,
            'search_trends_itch': np.float64,
            'search_trends_jaundice': np.float64,
            'search_trends_kidney_failure': np.float64,
            'search_trends_kidney_stone': np.float64,
            'search_trends_knee_pain': np.float64,
            'search_trends_kyphosis': np.float64,
            'search_trends_lactose_intolerance': np.float64,
            'search_trends_laryngitis': np.float64,
            'search_trends_leg_cramps': np.float64,
            'search_trends_lesion': np.float64,
            'search_trends_leukorrhea': np.float64,
            'search_trends_lightheadedness': np.float64,
            'search_trends_low_back_pain': np.float64,
            'search_trends_low_grade_fever': np.float64,
            'search_trends_lymphedema': np.float64,
            'search_trends_major_depressive_disorder': np.float64,
            'search_trends_malabsorption': np.float64,
            'search_trends_male_infertility': np.float64,
            'search_trends_manic_disorder': np.float64,
            'search_trends_melasma': np.float64,
            'search_trends_melena': np.float64,
            'search_trends_meningitis': np.float64,
            'search_trends_menorrhagia': np.float64,
            'search_trends_middle_back_pain': np.float64,
            'search_trends_migraine': np.float64,
            'search_trends_milium': np.float64,
            'search_trends_mitral_insufficiency': np.float64,
            'search_trends_mood_disorder': np.float64,
            'search_trends_mood_swing': np.float64,
            'search_trends_morning_sickness': np.float64,
            'search_trends_motion_sickness': np.float64,
            'search_trends_mouth_ulcer': np.float64,
            'search_trends_muscle_atrophy': np.float64,
            'search_trends_muscle_weakness': np.float64,
            'search_trends_myalgia': np.float64,
            'search_trends_mydriasis': np.float64,
            'search_trends_myocardial_infarction': np.float64,
            'search_trends_myoclonus': np.float64,
            'search_trends_nasal_congestion': np.float64,
            'search_trends_nasal_polyp': np.float64,
            'search_trends_nausea': np.float64,
            'search_trends_neck_mass': np.float64,
            'search_trends_neck_pain': np.float64,
            'search_trends_neonatal_jaundice': np.float64,
            'search_trends_nerve_injury': np.float64,
            'search_trends_neuralgia': np.float64,
            'search_trends_neutropenia': np.float64,
            'search_trends_night_sweats': np.float64,
            'search_trends_night_terror': np.float64,
            'search_trends_nocturnal_enuresis': np.float64,
            'search_trends_nodule': np.float64,
            'search_trends_nosebleed': np.float64,
            'search_trends_nystagmus': np.float64,
            'search_trends_obesity': np.float64,
            'search_trends_onychorrhexis': np.float64,
            'search_trends_oral_candidiasis': np.float64,
            'search_trends_orthostatic_hypotension': np.float64,
            'search_trends_osteopenia': np.float64,
            'search_trends_osteophyte': np.float64,
            'search_trends_osteoporosis': np.float64,
            'search_trends_otitis': np.float64,
            'search_trends_otitis_externa': np.float64,
            'search_trends_otitis_media': np.float64,
            'search_trends_pain': np.float64,
            'search_trends_palpitations': np.float64,
            'search_trends_pancreatitis': np.float64,
            'search_trends_panic_attack': np.float64,
            'search_trends_papule': np.float64,
            'search_trends_paranoia': np.float64,
            'search_trends_paresthesia': np.float64,
            'search_trends_pelvic_inflammatory_disease': np.float64,
            'search_trends_pericarditis': np.float64,
            'search_trends_periodontal_disease': np.float64,
            'search_trends_periorbital_puffiness': np.float64,
            'search_trends_peripheral_neuropathy': np.float64,
            'search_trends_perspiration': np.float64,
            'search_trends_petechia': np.float64,
            'search_trends_phlegm': np.float64,
            'search_trends_photodermatitis': np.float64,
            'search_trends_photophobia': np.float64,
            'search_trends_photopsia': np.float64,
            'search_trends_pleural_effusion': np.float64,
            'search_trends_pleurisy': np.float64,
            'search_trends_pneumonia': np.float64,
            'search_trends_podalgia': np.float64,
            'search_trends_polycythemia': np.float64,
            'search_trends_polydipsia': np.float64,
            'search_trends_polyneuropathy': np.float64,
            'search_trends_polyuria': np.float64,
            'search_trends_poor_posture': np.float64,
            'search_trends_post_nasal_drip': np.float64,
            'search_trends_postural_orthostatic_tachycardia_syndrome': (
                np.float64
            ),
            'search_trends_prediabetes': np.float64,
            'search_trends_proteinuria': np.float64,
            'search_trends_pruritus_ani': np.float64,
            'search_trends_psychosis': np.float64,
            'search_trends_ptosis': np.float64,
            'search_trends_pulmonary_edema': np.float64,
            'search_trends_pulmonary_hypertension': np.float64,
            'search_trends_purpura': np.float64,
            'search_trends_pus': np.float64,
            'search_trends_pyelonephritis': np.float64,
            'search_trends_radiculopathy': np.float64,
            'search_trends_rectal_pain': np.float64,
            'search_trends_rectal_prolapse': np.float64,
            'search_trends_red_eye': np.float64,
            'search_trends_renal_colic': np.float64,
            'search_trends_restless_legs_syndrome': np.float64,
            'search_trends_rheum': np.float64,
            'search_trends_rhinitis': np.float64,
            'search_trends_rhinorrhea': np.float64,
            'search_trends_rosacea': np.float64,
            'search_trends_round_ligament_pain': np.float64,
            'search_trends_rumination': np.float64,
            'search_trends_scar': np.float64,
            'search_trends_sciatica': np.float64,
            'search_trends_scoliosis': np.float64,
            'search_trends_seborrheic_dermatitis': np.float64,
            'search_trends_self_harm': np.float64,
            'search_trends_sensitivity_to_sound': np.float64,
            'search_trends_sexual_dysfunction': np.float64,
            'search_trends_shallow_breathing': np.float64,
            'search_trends_sharp_pain': np.float64,
            'search_trends_shivering': np.float64,
            'search_trends_shortness_of_breath': np.float64,
            'search_trends_shyness': np.float64,
            'search_trends_sinusitis': np.float64,
            'search_trends_skin_condition': np.float64,
            'search_trends_skin_rash': np.float64,
            'search_trends_skin_tag': np.float64,
            'search_trends_skin_ulcer': np.float64,
            'search_trends_sleep_apnea': np.float64,
            'search_trends_sleep_deprivation': np.float64,
            'search_trends_sleep_disorder': np.float64,
            'search_trends_snoring': np.float64,
            'search_trends_sore_throat': np.float64,
            'search_trends_spasticity': np.float64,
            'search_trends_splenomegaly': np.float64,
            'search_trends_sputum': np.float64,
            'search_trends_stomach_rumble': np.float64,
            'search_trends_strabismus': np.float64,
            'search_trends_stretch_marks': np.float64,
            'search_trends_stridor': np.float64,
            'search_trends_stroke': np.float64,
            'search_trends_stuttering': np.float64,
            'search_trends_subdural_hematoma': np.float64,
            'search_trends_suicidal_ideation': np.float64,
            'search_trends_swelling': np.float64,
            'search_trends_swollen_feet': np.float64,
            'search_trends_swollen_lymph_nodes': np.float64,
            'search_trends_syncope': np.float64,
            'search_trends_tachycardia': np.float64,
            'search_trends_tachypnea': np.float64,
            'search_trends_telangiectasia': np.float64,
            'search_trends_tenderness': np.float64,
            'search_trends_testicular_pain': np.float64,
            'search_trends_throat_irritation': np.float64,
            'search_trends_thrombocytopenia': np.float64,
            'search_trends_thyroid_nodule': np.float64,
            'search_trends_tic': np.float64,
            'search_trends_tinnitus': np.float64,
            'search_trends_tonsillitis': np.float64,
            'search_trends_toothache': np.float64,
            'search_trends_tremor': np.float64,
            'search_trends_trichoptilosis': np.float64,
            'search_trends_tumor': np.float64,
            'search_trends_type_2_diabetes': np.float64,
            'search_trends_unconsciousness': np.float64,
            'search_trends_underweight': np.float64,
            'search_trends_upper_respiratory_tract_infection': np.float64,
            'search_trends_urethritis': np.float64,
            'search_trends_urinary_incontinence': np.float64,
            'search_trends_urinary_tract_infection': np.float64,
            'search_trends_urinary_urgency': np.float64,
            'search_trends_uterine_contraction': np.float64,
            'search_trends_vaginal_bleeding': np.float64,
            'search_trends_vaginal_discharge': np.float64,
            'search_trends_vaginitis': np.float64,
            'search_trends_varicose_veins': np.float64,
            'search_trends_vasculitis': np.float64,
            'search_trends_ventricular_fibrillation': np.float64,
            'search_trends_ventricular_tachycardia': np.float64,
            'search_trends_vertigo': np.float64,
            'search_trends_viral_pneumonia': np.float64,
            'search_trends_visual_acuity': np.float64,
            'search_trends_vomiting': np.float64,
            'search_trends_wart': np.float64,
            'search_trends_water_retention': np.float64,
            'search_trends_weakness': np.float64,
            'search_trends_weight_gain': np.float64,
            'search_trends_wheeze': np.float64,
            'search_trends_xeroderma': np.float64,
            'search_trends_xerostomia': np.float64,
            'search_trends_yawn': np.float64,
            'new_recovered_age_0': np.float64,
            'new_recovered_age_1': np.float64,
            'new_recovered_age_2': np.float64,
            'new_recovered_age_3': np.float64,
            'new_recovered_age_4': np.float64,
            'new_recovered_age_5': np.float64,
            'new_recovered_age_6': np.float64,
            'new_recovered_age_7': np.float64,
            'new_recovered_age_8': np.float64,
            'new_recovered_age_9': np.float64,
            'cumulative_recovered_age_0': np.float64,
            'cumulative_recovered_age_1': np.float64,
            'cumulative_recovered_age_2': np.float64,
            'cumulative_recovered_age_3': np.float64,
            'cumulative_recovered_age_4': np.float64,
            'cumulative_recovered_age_5': np.float64,
            'cumulative_recovered_age_6': np.float64,
            'cumulative_recovered_age_7': np.float64,
            'cumulative_recovered_age_8': np.float64,
            'cumulative_recovered_age_9': np.float64,
            'new_recovered_male': np.float64,
            'new_recovered_female': np.float64,
            'cumulative_recovered_male': np.float64,
            'cumulative_recovered_female': np.float64,
            'locality_code': np.str_,
            'locality_name': np.str_,
            'new_ventilator_patients': np.float64,
            'cumulative_ventilator_patients': np.float64,
            'new_persons_fully_vaccinated_pfizer': np.float64,
            'cumulative_persons_fully_vaccinated_pfizer': np.float64,
            'new_vaccine_doses_administered_pfizer': np.float64,
            'cumulative_vaccine_doses_administered_pfizer': np.float64,
            'new_persons_fully_vaccinated_moderna': np.float64,
            'cumulative_persons_fully_vaccinated_moderna': np.float64,
            'new_vaccine_doses_administered_moderna': np.float64,
            'cumulative_vaccine_doses_administered_moderna': np.float64,
            'new_persons_fully_vaccinated_janssen': np.float64,
            'cumulative_persons_fully_vaccinated_janssen': np.float64,
            'new_vaccine_doses_administered_janssen': np.float64,
            'cumulative_vaccine_doses_administered_janssen': np.float64,
        }),
        supervised_keys=None,
        homepage='https://github.com/GoogleCloudPlatform/covid-19-open-data',
        citation=_CITATION,
    )

  def _split_generators(self, dl_manager: tfds.download.DownloadManager):
    """Returns SplitGenerators."""
    archive_path = dl_manager.download(
        'https://storage.googleapis.com/covid19-open-data/v3/aggregated.csv.gz?generation=1620814656792419'
    )
    return {tfds.Split.TRAIN: self._generate_examples(dl_manager, archive_path)}

  def _generate_examples(
      self, dl_manager: tfds.download.DownloadManager, archive_path
  ):
    """Yields examples."""
    pd = tfds.core.lazy_imports.pandas
    beam = tfds.core.lazy_imports.apache_beam

    def cast_according_to_column(feature_type, v):
      if feature_type == tf.string and isinstance(v, (float, int)):
        return str(v)
      return v

    file_handles = dl_manager.iter_archive(archive_path)
    _, file = next(file_handles)

    columns = pd.read_csv(file, nrows=1).columns

    def load_shard(index: int):
      # There is only one file so by using the for we guarantee that the file
      # will be closed.
      for _, file in dl_manager.iter_archive(archive_path):
        df = pd.read_csv(file, skiprows=index, nrows=_BATCH_SIZE)
        features = self.info.features
        result = []
        for i, row in df.iterrows():
          example = {
              k: cast_according_to_column(features[k].dtype, v)
              for k, v in zip(columns, row.values)
          }
          result.append((index + i, example))
        return result

    return beam.Create(list(range(0, _N_RECORDS, _BATCH_SIZE))) | beam.FlatMap(
        load_shard
    )



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""robosuite_panda_pick_place_can dataset."""
import sys

import pytest
import tensorflow_datasets.public_api as tfds
from tensorflow_datasets.rlds.robosuite_panda_pick_place_can import robosuite_panda_pick_place_can


pytest.importorskip(
    'envlogger',
    reason=(
        '`envlogger` library might not be available for Python'
        f' {sys.version} or platform {sys.platform}; see'
        ' https://pypi.org/project/envlogger/#files'
    ),
)


class RobosuitePandaPickPlaceCanHumanTest(tfds.testing.DatasetBuilderTestCase):
  """Tests for robosuite_panda_pick_place_can dataset."""

  DATASET_CLASS = robosuite_panda_pick_place_can.RobosuitePandaPickPlaceCan
  SPLITS = {
      'train': 2,  # Number of fake train example
  }

  SKIP_TF1_GRAPH_MODE = True

  BUILDER_CONFIG_NAMES_TO_TEST = ['human_dc29b40a']

  DL_EXTRACT_RESULT = {'file_path': 'human_dc29b40a'}
  DL_DOWNLOAD_RESULT = {'file_path': 'human_dc29b40a'}


class RobosuitePandaPickPlaceCanHumanImagesTest(
    tfds.testing.DatasetBuilderTestCase
):
  """Tests for robosuite_panda_pick_place_can dataset."""

  DATASET_CLASS = robosuite_panda_pick_place_can.RobosuitePandaPickPlaceCan
  SPLITS = {
      'train': 2,  # Number of fake train example
  }

  SKIP_TF1_GRAPH_MODE = True

  BUILDER_CONFIG_NAMES_TO_TEST = ['human_images_dc29b40a']

  DL_EXTRACT_RESULT = {'file_path': 'human_images_dc29b40a'}
  DL_DOWNLOAD_RESULT = {'file_path': 'human_images_dc29b40a'}


class RobosuitePandaPickPlaceCanSyntheticTest(
    tfds.testing.DatasetBuilderTestCase
):
  """Tests for robosuite_panda_pick_place_can dataset."""

  DATASET_CLASS = robosuite_panda_pick_place_can.RobosuitePandaPickPlaceCan
  SPLITS = {
      'train': 2,  # Number of fake train example
  }

  SKIP_TF1_GRAPH_MODE = True

  BUILDER_CONFIG_NAMES_TO_TEST = ['synthetic_stochastic_sac_afe13968']

  DL_EXTRACT_RESULT = {'file_path': 'synthetic_stochastic_sac_afe13968'}
  DL_DOWNLOAD_RESULT = {'file_path': 'synthetic_stochastic_sac_afe13968'}


if __name__ == '__main__':
  tfds.testing.test_main()



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Dataset collection catalog documentation template."""

import abc
import textwrap
from typing import List

import tensorflow_datasets as tfds
from tensorflow_datasets.scripts.documentation import dataset_markdown_builder as dmb
from tensorflow_datasets.scripts.documentation import doc_utils

# Token used to indicate the section shouldn't be displayed
_SKIP_SECTION = object()


class CollectionSection(abc.ABC):
  """Abstract class for a documentation Section (description, homepage, ...)."""

  NAME: str
  EXTRA_DOC: str = ''

  @abc.abstractmethod
  def get_key(self, loader: tfds.core.DatasetCollectionLoader) -> dmb.Key:
    """Get the key of the section.

    The key is used to merge similar sections across builder configs. For
    instance, https://www.tensorflow.org/datasets/catalog/wiki40b only display
    once the `FeatureDict`, homepage,... as those sections are the same for
    all configs.

    Args:
      loader: The loader of the DatasetCollection to document.

    Returns:
      key: The section key.
    """
    pass

  @abc.abstractmethod
  def content(self, loader: tfds.core.DatasetCollectionLoader) -> str:
    """Returns the section content."""
    pass

  def display(self, loader: tfds.core.DatasetCollectionLoader) -> str:
    """Returns the section content."""
    content = self.content(loader)
    if content is _SKIP_SECTION:
      return ''
    header = f'*   **{self.NAME}**{self.EXTRA_DOC}'
    is_block = isinstance(content, dmb.Block)
    if not isinstance(content, dmb.IntentedBlock):
      content = content.strip()  # Note: `strip()` cast `Block` -> `str`
    if is_block:
      content = f'{header}:\n\n{content}\n\n'
    else:
      content = f'{header}: {content}\n\n'
    return content


# ------------------------------ Sections -------------------------------------

# pylint:disable = missing-class-docstring


class CollectionHomepageSection(CollectionSection):
  NAME = 'Homepage'

  def get_key(self, loader: tfds.core.DatasetCollectionLoader):
    return loader.collection.info.homepage

  def content(self, loader: tfds.core.DatasetCollectionLoader):
    homepage = self.get_key(loader)
    if homepage:
      homepage = doc_utils.format_homepage_url(homepage)
    else:
      homepage = ''
    return dmb.Block(homepage)


class CollectionDescriptionSection(CollectionSection):
  NAME = 'Description'

  def get_key(self, loader: tfds.core.DatasetCollectionLoader):
    return loader.collection.info.description

  def content(self, loader: tfds.core.DatasetCollectionLoader):
    return dmb.Block(loader.collection.info.description)


class CollectionVersionSection(CollectionSection):
  """Lists versions in the dataset collection, along with their datasets."""

  NAME = 'Versions'

  def _list_versions(self, loader: tfds.core.DatasetCollectionLoader):
    """List versions."""
    # Default version is the latest.
    default_version = loader.collection.get_latest_version()
    all_versions = [
        *loader.collection.info.release_notes,
        *loader.collection.datasets,
    ]

    all_versions_and_notes = loader.collection.info.release_notes

    all_versions = set(tfds.core.Version(v) for v in all_versions)

    for v in sorted(all_versions):  # List all available versions.
      if v == default_version:  # Highlight the default version.
        version_name = '**`{}`** (default)'.format(str(v))
      else:
        version_name = '`{}`'.format(str(v))
      description = all_versions_and_notes.get(str(v), 'No release notes.')
      yield f'    * {version_name}: {description}'

  def get_key(self, loader: tfds.core.DatasetCollectionLoader):
    release_key = tuple(
        (k, v) for k, v in loader.collection.info.release_notes.items()
    )
    return (tuple(loader.collection.all_versions), release_key)

  def content(self, loader: tfds.core.DatasetCollectionLoader):
    return dmb.IntentedBlock('\n'.join(self._list_versions(loader)))


class CollectionDatasetsSection(CollectionSection):
  NAME = 'Datasets in the default version'

  def get_key(self, loader: tfds.core.DatasetCollectionLoader):
    return loader.collection.get_latest_version()

  def _get_dataset_catalog_link(self, reference):
    """Returns the link to a dataset documentation in TFDS catalog."""
    prefix = 'https://www.tensorflow.org/datasets/catalog/'
    # Links to TFDS catalog's entries don't include versions.
    ds_link = prefix + reference.dataset_name
    # For configs, return precise link to config documentation.
    if reference.config:
      ds_link += f'#{reference.config}'
    return ds_link

  def _list_datasets(
      self, collection_loader: tfds.core.DatasetCollectionLoader
  ):
    """List datasets for the latest version."""
    # get_collection retrieves the datasets in the default (= latest) version.
    datasets = collection_loader.collection.get_collection()
    for name, reference in datasets.items():
      ds_link = self._get_dataset_catalog_link(reference)
      yield f'    * `{name}`: [`{reference.tfds_name()}`]({ds_link})'

  def content(self, loader: tfds.core.DatasetCollectionLoader):
    return dmb.IntentedBlock('\n'.join(self._list_datasets(loader)))


class CollectionCitationSection(CollectionSection):
  NAME = 'Citation'

  def get_key(self, loader: tfds.core.DatasetCollectionLoader):
    return loader.collection.info.citation

  def content(self, loader: tfds.core.DatasetCollectionLoader):
    if not loader.collection.info.citation:
      return ''
    return dmb.Block(
        textwrap.dedent(
            f"""
            ```
            {tfds.core.utils.indent(loader.collection.info.citation, '            ')}
            ```
            """
        )
    )


# --------------------------- Main page ---------------------------


def _display_collection_heading(
    collection: tfds.core.DatasetCollectionLoader,
) -> str:
  return f"""
      # `{collection.collection_name}`

      """


def _display_collection_sections(
    loader: tfds.core.DatasetCollectionLoader,
    all_sections: List[CollectionSection],
) -> str:
  return ''.join([section.display(loader) for section in all_sections])


def get_collection_markdown_string(
    *,
    collection: tfds.core.DatasetCollectionLoader,
) -> str:
  """Build the collection markdown."""
  all_sections = [
      CollectionDescriptionSection(),
      CollectionHomepageSection(),
      CollectionVersionSection(),
      CollectionDatasetsSection(),
      CollectionCitationSection(),
  ]

  doc_str = [
      _display_collection_heading(collection),
      _display_collection_sections(
          loader=collection, all_sections=all_sections
      ),
  ]
  return '\n\n'.join([tfds.core.utils.dedent(s) for s in doc_str if s])



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: skip-file

# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: tensorflow_datasets/proto/tf_feature.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n*tensorflow_datasets/proto/tf_feature.proto\x12\x0ftensorflow_copy"\x1a\n\tBytesList\x12\r\n\x05value\x18\x01'
    b' \x03(\x0c"\x1e\n\tFloatList\x12\x11\n\x05value\x18\x01'
    b' \x03(\x02\x42\x02\x10\x01"\x1e\n\tInt64List\x12\x11\n\x05value\x18\x01'
    b' \x03(\x03\x42\x02\x10\x01"\xa7\x01\n\x07\x46\x65\x61ture\x12\x30\n\nbytes_list\x18\x01'
    b' \x01(\x0b\x32\x1a.tensorflow_copy.BytesListH\x00\x12\x30\n\nfloat_list\x18\x02'
    b' \x01(\x0b\x32\x1a.tensorflow_copy.FloatListH\x00\x12\x30\n\nint64_list\x18\x03'
    b' \x01(\x0b\x32\x1a.tensorflow_copy.Int64ListH\x00\x42\x06\n\x04kind"\x8d\x01\n\x08\x46\x65\x61tures\x12\x37\n\x07\x66\x65\x61ture\x18\x01'
    b' \x03(\x0b\x32&.tensorflow_copy.Features.FeatureEntry\x1aH\n\x0c\x46\x65\x61tureEntry\x12\x0b\n\x03key\x18\x01'
    b" \x01(\t\x12'\n\x05value\x18\x02"
    b' \x01(\x0b\x32\x18.tensorflow_copy.Feature:\x02\x38\x01"8\n\x0b\x46\x65\x61tureList\x12)\n\x07\x66\x65\x61ture\x18\x01'
    b' \x03(\x0b\x32\x18.tensorflow_copy.Feature"\xa6\x01\n\x0c\x46\x65\x61tureLists\x12\x44\n\x0c\x66\x65\x61ture_list\x18\x01'
    b' \x03(\x0b\x32..tensorflow_copy.FeatureLists.FeatureListEntry\x1aP\n\x10\x46\x65\x61tureListEntry\x12\x0b\n\x03key\x18\x01'
    b' \x01(\t\x12+\n\x05value\x18\x02'
    b' \x01(\x0b\x32\x1c.tensorflow_copy.FeatureList:\x02\x38\x01\x42\x81\x01\n\x16org.tensorflow.exampleB\rFeatureProtosP\x01ZSgithub.com/tensorflow/tensorflow/tensorflow/go/core/example/example_protos_go_proto\xf8\x01\x01\x62\x06proto3'
)

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(
    DESCRIPTOR, 'tensorflow_datasets.proto.tf_feature_pb2', globals()
)
if _descriptor._USE_C_DESCRIPTORS == False:
  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\026org.tensorflow.exampleB\rFeatureProtosP\001ZSgithub.com/tensorflow/tensorflow/tensorflow/go/core/example/example_protos_go_proto\370\001\001'
  _FLOATLIST.fields_by_name['value']._options = None
  _FLOATLIST.fields_by_name['value']._serialized_options = b'\020\001'
  _INT64LIST.fields_by_name['value']._options = None
  _INT64LIST.fields_by_name['value']._serialized_options = b'\020\001'
  _FEATURES_FEATUREENTRY._options = None
  _FEATURES_FEATUREENTRY._serialized_options = b'8\001'
  _FEATURELISTS_FEATURELISTENTRY._options = None
  _FEATURELISTS_FEATURELISTENTRY._serialized_options = b'8\001'
  _BYTESLIST._serialized_start = 63
  _BYTESLIST._serialized_end = 89
  _FLOATLIST._serialized_start = 91
  _FLOATLIST._serialized_end = 121
  _INT64LIST._serialized_start = 123
  _INT64LIST._serialized_end = 153
  _FEATURE._serialized_start = 156
  _FEATURE._serialized_end = 323
  _FEATURES._serialized_start = 326
  _FEATURES._serialized_end = 467
  _FEATURES_FEATUREENTRY._serialized_start = 395
  _FEATURES_FEATUREENTRY._serialized_end = 467
  _FEATURELIST._serialized_start = 469
  _FEATURELIST._serialized_end = 525
  _FEATURELISTS._serialized_start = 528
  _FEATURELISTS._serialized_end = 694
  _FEATURELISTS_FEATURELISTENTRY._serialized_start = 614
  _FEATURELISTS_FEATURELISTENTRY._serialized_end = 694
# @@protoc_insertion_point(module_scope)



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Davis dataset test."""

import numpy as np
import tensorflow as tf
import tensorflow_datasets.public_api as tfds
from tensorflow_datasets.video.davis import davis


class DavisTest(tfds.testing.DatasetBuilderTestCase):
  """Tests for davis dataset."""

  BUILDER_CONFIG_NAMES_TO_TEST = ['480p', 'full_resolution']
  DATASET_CLASS = davis.Davis
  SPLITS = {
      tfds.Split.TRAIN: 1,  # Number of fake train examples.
      tfds.Split.VALIDATION: 1,
  }

  def _download_and_prepare_as_dataset(self, builder):
    super()._download_and_prepare_as_dataset(builder)

    if not tf.executing_eagerly():  # Only test the following in eager mode.
      return

    with self.subTest('annotations_classes'):
      splits = builder.as_dataset()
      train_ex = list(splits[tfds.Split.TRAIN])[0]
      val_ex = list(splits[tfds.Split.VALIDATION])[0]
      # Check that the dataset examples contain the correct number of classes.
      self.assertSetEqual(
          set(np.unique(train_ex['video']['segmentations'].numpy())), {0, 1}
      )
      self.assertSetEqual(
          set(np.unique(val_ex['video']['segmentations'].numpy())), {0, 1, 2}
      )

    with self.subTest('dataset_shapes'):
      splits = builder.as_dataset()
      train_ex = list(splits[tfds.Split.TRAIN])[0]
      val_ex = list(splits[tfds.Split.VALIDATION])[0]
      train_num_frames = train_ex['metadata']['num_frames'].numpy()
      val_num_frames = val_ex['metadata']['num_frames'].numpy()
      # Check that the shapes of the dataset examples is correct.
      self.assertEqual(train_num_frames, 4)
      self.assertEqual(val_num_frames, 4)
      _, height, width, _ = train_ex['video']['frames'].numpy().shape
      self.assertEqual(
          train_ex['video']['frames'].numpy().shape, (4, height, width, 3)
      )
      self.assertEqual(
          train_ex['video']['segmentations'].numpy().shape,
          (4, height, width, 1),
      )
      _, height, width, _ = val_ex['video']['frames'].numpy().shape
      self.assertEqual(
          val_ex['video']['frames'].numpy().shape, (4, height, width, 3)
      )
      self.assertEqual(
          val_ex['video']['segmentations'].numpy().shape, (4, height, width, 1)
      )


if __name__ == '__main__':
  tfds.testing.test_main()



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utils for ConllDatasetBuilder.

This file contains list of reusable tags which are commonly used in CoNLL-based
datasets.
"""
import collections

from tensorflow_datasets.core.dataset_builders.conll import conll_dataset_builder
from tensorflow_datasets.core.features.class_label_feature import ClassLabel
from tensorflow_datasets.core.features.sequence_feature import Sequence
from tensorflow_datasets.core.features.text_feature import Text

# Standard POS tags for family of CoNLL-based datasets.

# Used by: conll2003
POS_TAGS = (
    '"',
    "''",
    "#",
    "$",
    "(",
    ")",
    ",",
    ".",
    ":",
    "``",
    "CC",
    "CD",
    "DT",
    "EX",
    "FW",
    "IN",
    "JJ",
    "JJR",
    "JJS",
    "LS",
    "MD",
    "NN",
    "NNP",
    "NNPS",
    "NNS",
    "NN|SYM",
    "PDT",
    "POS",
    "PRP",
    "PRP$",
    "RB",
    "RBR",
    "RBS",
    "RP",
    "SYM",
    "TO",
    "UH",
    "VB",
    "VBD",
    "VBG",
    "VBN",
    "VBP",
    "VBZ",
    "WDT",
    "WP",
    "WP$",
    "WRB",
)

# Standard chunk tags for family of CoNLL-based datasets.

# Used by: conll2003
CHUNK_TAGS = (
    "O",
    "B-ADJP",
    "I-ADJP",
    "B-ADVP",
    "I-ADVP",
    "B-CONJP",
    "I-CONJP",
    "B-INTJ",
    "I-INTJ",
    "B-LST",
    "I-LST",
    "B-NP",
    "I-NP",
    "B-PP",
    "I-PP",
    "B-PRT",
    "I-PRT",
    "B-SBAR",
    "I-SBAR",
    "B-UCP",
    "I-UCP",
    "B-VP",
    "I-VP",
)

# Standard NER tags for family of CoNLL-based datasets.

# Used by: conll2002, conll2003
NER_TAGS = (
    "O",
    "B-PER",
    "I-PER",
    "B-ORG",
    "I-ORG",
    "B-LOC",
    "I-LOC",
    "B-MISC",
    "I-MISC",
)

# Standard separators for family of CoNLL-based datasets.

# Used by: conll2002, conll2003
CONLL_2002_SEPARATOR = " "

# Standard orderedfeature dicts for family of CoNLL-based datasets.
# Importantly, the order of the features should map the order of the feature
# columns in the input files.

# Used by: conll2003
CONLL_2003_ORDERED_FEATURES = collections.OrderedDict({
    "tokens": Sequence(Text()),
    "pos": Sequence(ClassLabel(names=POS_TAGS)),
    "chunks": Sequence(ClassLabel(names=CHUNK_TAGS)),
    "ner": Sequence(ClassLabel(names=NER_TAGS)),
})

# Standard ConllBuilderConfig for family of CoNLL-based datasets.

# Used by: conll2003
CONLL_2003_CONFIG = conll_dataset_builder.ConllBuilderConfig(
    name="conll2003",
    separator=CONLL_2002_SEPARATOR,
    ordered_features=CONLL_2003_ORDERED_FEATURES,
)



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utilities for file names."""

from __future__ import annotations

import dataclasses
import functools
import os
import re
import textwrap
from typing import Any, Dict, List, Mapping, MutableMapping, Optional, Tuple, Union

from etils import epath
from tensorflow_datasets.core.utils import py_utils
from tensorflow_datasets.core.utils import version as version_lib

_NAME_CLASS = r'[a-zA-Z][\w]*'
_NAME_CLASS_REG = re.compile(r'^' + _NAME_CLASS + r'$')

# Regex matching 'dataset/config:1.*.*/arg=123'
_NAME_REG = re.compile(
    r'^'
    r'(?P<dataset_name>([\w\-]+:)?' + _NAME_CLASS + r')'
    r'(/(?P<config>[\w\+\-\.]+))?'
    r'(:(?P<version>(\d+|\*)(\.(\d+|\*)){2}))?'
    r'(/(?P<kwargs>(\w+=\w+)(,\w+=[^,]+)*))?'
    r'$'
)

_DEFAULT_NUM_DIGITS_FOR_SHARDS = 5

_VAR_DATASET = 'DATASET'
_VAR_SPLIT = 'SPLIT'
_VAR_SHARD_INDEX = 'SHARD_INDEX'
_VAR_NUM_SHARDS = 'NUM_SHARDS'
_VAR_SHARD_X_OF_Y = 'SHARD_X_OF_Y'
_VAR_FILEFORMAT = 'FILEFORMAT'
_VAR_REGEX_MAPPING = {
    _VAR_DATASET: rf'(?P<dataset_name>{_NAME_CLASS})',
    _VAR_FILEFORMAT: r'(?P<filetype_suffix>\w+)',
    _VAR_SPLIT: r'(?P<split>(\w|-)+)',
    _VAR_SHARD_INDEX: r'(?P<shard_index>\d{5,})',
    _VAR_NUM_SHARDS: r'(?P<num_shards>\d{5,})',
    _VAR_SHARD_X_OF_Y: r'(?P<shard_index>\d{5,})-of-(?P<num_shards>\d{5,})',
}

DEFAULT_FILENAME_TEMPLATE = '{DATASET}-{SPLIT}.{FILEFORMAT}-{SHARD_X_OF_Y}'

_first_cap_re = re.compile('(.)([A-Z][a-z0-9]+)')
_all_cap_re = re.compile('([a-z0-9])([A-Z])')

Value = Union[str, int, float, bool]


@dataclasses.dataclass(eq=True, order=True, frozen=True)
class DatasetName:
  """Dataset namespace+name."""

  namespace: Optional[str]
  name: str

  def __init__(
      self,
      namespace_name: Optional[str] = None,
      *,
      namespace: Optional[str] = None,
      name: Optional[str] = None,
  ):
    if namespace_name and bool(namespace or name):
      raise ValueError(
          "Name should be defined by `DatasetName('ns:name')` or "
          "`DatasetName(namespace='ns', name='name'). Mixing args and kwargs "
          'is invalid.'
      )
    if namespace_name:
      if ':' in namespace_name:
        namespace, name = namespace_name.split(':')
      else:
        namespace, name = None, namespace_name
    super().__setattr__('namespace', namespace)
    super().__setattr__('name', name)

  def __str__(self) -> str:
    return f'{self.namespace}:{self.name}' if self.namespace else self.name

  def __repr__(self) -> str:
    return f'{type(self).__name__}({str(self)!r})'


def is_valid_dataset_name(name_str: str) -> bool:
  """Returns True is the given name is a valid dataset name."""
  res = _NAME_REG.match(name_str)
  return bool(res)


def is_valid_dataset_and_class_name(name_str: str) -> bool:
  """Returns True is the given name is a valid dataset and class name."""
  res = _NAME_CLASS_REG.match(name_str)
  return bool(res)


def parse_builder_name_kwargs(
    name: str,
    **builder_kwargs: Any,
) -> Tuple[DatasetName, Dict[str, Any]]:
  """Normalize builder kwargs.

  Example:

  ```python
  ds_name, builder_kwargs = parse_builder_name_kwargs(
      'kaggle:ds/cfg:1.2.3', data_dir='...'
  )
  ds_name.namespace == 'kaggle'
  ds_name.name == 'ds'
  builder_kwargs == {'config': 'cfg', 'version': '1.2.3', 'data_dir': '...'}
  ```

  Args:
    name: Name
    **builder_kwargs: Builder kwargs

  Returns:
    ns_name: DatasetName object for the given dataset name
    builder_kwargs: Builder kwargs (version, config, data_dir,...)
  """
  name, parsed_builder_kwargs = _dataset_name_and_kwargs_from_name_str(name)
  builder_kwargs = dict(**parsed_builder_kwargs, **builder_kwargs)
  return DatasetName(name), builder_kwargs


def _dataset_name_and_kwargs_from_name_str(
    name_str: str,
) -> Tuple[str, Dict[str, Value]]:
  """Extract kwargs from name str."""
  err_msg = textwrap.dedent(f"""\
      Parsing builder name string {name_str} failed.
      The builder name string must be of the following format:
        dataset_name[/config_name][:version][/kwargs]

        Where:

          * dataset_name and config_name are string following python variable naming.
          * version is of the form x.y.z where {{x,y,z}} can be any digit or *.
          * kwargs is a comma list separated of arguments and values to pass to
            builder.

        Examples:
          my_dataset
          my_dataset:1.2.*
          my_dataset/config1
          my_dataset/config1:1.*.*
          my_dataset/config1/arg1=val1,arg2=val2
          my_dataset/config1:1.2.3/right=True,foo=bar,rate=1.2
      """)

  res = _NAME_REG.match(name_str)
  if not res:
    raise ValueError(err_msg)
  name = res.group('dataset_name')
  # Normalize the name to accept CamelCase
  name = camelcase_to_snakecase(name)
  kwargs = _kwargs_str_to_kwargs(res.group('kwargs'))
  try:
    for attr in ['config', 'version']:
      val = res.group(attr)
      if val is None:
        continue
      if attr in kwargs:
        raise ValueError('Dataset %s: cannot pass %s twice.' % (name, attr))
      kwargs[attr] = val
    return name, kwargs
  except Exception as e:  # pylint: disable=broad-except
    py_utils.reraise(e, prefix=err_msg)  # pytype: disable=bad-return-type


@dataclasses.dataclass(order=True)
class DatasetReference:
  """Reference to a dataset.

  Attributes:
    dataset_name: name of the dataset.
    namespace: optional namespace in which this dataset belongs.
    config: optional config to be used in the dataset.
    version: version of the dataset to be used. If `None`, the latest version
      will be loaded. An error is raised if the specified version cannot be
      provided.
    data_dir: Optional data dir where this dataset is located. If None, defaults
      to the value of the environment variable TFDS_DATA_DIR, if set, otherwise
    split_mapping: mapping between split names. If the `DatasetCollection` wants
      to use different split names than the source datasets, then this mapping
      can be used. For example, if the collection uses the split `valid`, but
      this dataset uses the split `validation`, then the `split_mapping` should
      be `{'validation': 'valid'}`.
  """

  dataset_name: str
  namespace: None | str = None
  config: None | str = None
  version: None | str | version_lib.Version = None
  data_dir: None | str | os.PathLike = None  # pylint: disable=g-bare-generic
  split_mapping: None | Mapping[str, str] = None

  def __post_init__(self):
    if isinstance(self.version, version_lib.Version):
      self.version = str(self.version)

  def tfds_name(self, include_version: bool = True) -> str:
    """Returns the TFDS name of the referenced dataset.

    Args:
      include_version: whether to include the dataset version in the tfds name.
        For example, this would result in `dataset/config:1.0.0` if set to True,
        or in `dataset/config` if set to False. Default is True.

    Returns:
      The TFDS name of the `DatasetReference`.
    """
    dataset_name = self.dataset_name
    if self.namespace:
      dataset_name = f'{self.namespace}:{dataset_name}'
    if self.config:
      dataset_name += f'/{self.config}'
    if self.version and include_version:
      dataset_name += f':{self.version}'
    return dataset_name

  def get_split(self, split: str) -> str:
    if self.split_mapping:
      return self.split_mapping.get(split, split)
    return split

  def dataset_dir(
      self,
      data_dir: Optional[epath.PathLike] = None,
  ) -> epath.Path:
    """Returns the path where the data of this dataset lives.

    Example: `/my_data_dir/datasets/c4/en/3.0.0`.

    Arguments:
      data_dir: optional path where this dataset is stored. If not specified,
        then it uses the `data_dir` specified in this dataset reference. If that
        is not specified either, then a `ValueError` is raised.

    Returns:
      the path where the data of this dataset lives.
    """
    data_dir = data_dir or self.data_dir
    if data_dir is None:
      raise ValueError('No data dir was specified!')
    dataset_dir: epath.Path = epath.Path(data_dir) / self.dataset_name
    if self.config:
      dataset_dir = dataset_dir / self.config
    if self.version is None:
      raise ValueError(
          "Version wasn't specified and is needed to get the dataset dir!"
      )
    dataset_dir = dataset_dir / str(self.version)
    return dataset_dir

  def replace(self, **kwargs: Any) -> DatasetReference:
    """Returns a copy with updated attributes."""
    return dataclasses.replace(self, **kwargs)

  @classmethod
  def from_tfds_name(
      cls,
      tfds_name: str,
      split_mapping: Optional[Mapping[str, str]] = None,
      data_dir: Union[None, str, os.PathLike] = None,  # pylint: disable=g-bare-generic
  ) -> DatasetReference:
    """Returns the `DatasetReference` for the given TFDS dataset."""
    parsed_name, builder_kwargs = parse_builder_name_kwargs(tfds_name)
    version, config = None, None
    version = builder_kwargs.get('version')
    config = builder_kwargs.get('config')
    return cls(
        dataset_name=parsed_name.name,
        namespace=parsed_name.namespace,
        version=version,
        config=config,
        split_mapping=split_mapping,
        data_dir=data_dir,
    )


def references_for(
    name_to_tfds_name: Mapping[str, str]
) -> Mapping[str, DatasetReference]:
  """Constructs of dataset references.

  Note that you can specify the config and the version in the TFDS name.
  For example:
  ```
  references_for(name_to_tfds_name={
    "wiki_it": "wikipedia/20201201.it:1.0.0",
    "wiki_en": "scan/length:1.1.1",
  })
  ```

  Args:
    name_to_tfds_name: The mapping between name to be used in the dataset
      collection and the TFDS name (plus optional config and version).

  Returns:
    Returns a dictionary of dataset_name: `DatasetReference`.
  """
  return {
      name: DatasetReference.from_tfds_name(tfds_name)
      for name, tfds_name in name_to_tfds_name.items()
  }


def reference_for(tfds_name: str) -> DatasetReference:
  """Returns the corresponding `DatasetReference` for a TFDS dataset name."""
  return DatasetReference.from_tfds_name(tfds_name)


def _kwargs_str_to_kwargs(kwargs_str: str):
  """Converts given `kwargs` as str into kwargs dict."""
  if not kwargs_str:
    return {}
  kwarg_strs = kwargs_str.split(',')
  kwargs = {}
  for kwarg_str in kwarg_strs:
    kwarg_name, kwarg_val = kwarg_str.split('=')
    kwargs[kwarg_name] = _cast_to_pod(kwarg_val)
  return kwargs


def _cast_to_pod(val: str) -> Value:
  """Try cast to bool, int, float, str, in that order."""
  bools = {'True': True, 'False': False}
  if val in bools:
    return bools[val]
  try:
    return int(val)
  except ValueError:
    try:
      return float(val)
    except ValueError:
      return val


def camelcase_to_snakecase(name: str) -> str:
  """Convert camel-case string to snake-case."""
  s1 = _first_cap_re.sub(r'\1_\2', name)
  return _all_cap_re.sub(r'\1_\2', s1).lower()


def snake_to_camelcase(name: str) -> str:
  """Convert snake-case string to camel-case string."""
  return ''.join(n.capitalize() for n in name.split('_'))


def filename_prefix_for_name(name: str) -> str:
  if os.path.basename(name) != name:
    raise ValueError('Should be a dataset name, not a path: %s' % name)
  return camelcase_to_snakecase(name)


def filename_prefix_for_split(name: str, split: str) -> str:
  if os.path.basename(name) != name:
    raise ValueError('Should be a dataset name, not a path: %s' % name)
  return '%s-%s' % (filename_prefix_for_name(name), split)


def _num_digits_needed(num_shards: Optional[int]) -> int:
  return max(len(str(num_shards or 0)), _DEFAULT_NUM_DIGITS_FOR_SHARDS)


def _replace_shard_suffix(filepath: str, replacement: str) -> str:
  """Replaces the shard suffix (must be at the end) with the given string."""
  (new_string, num_subs) = re.subn(
      pattern=r'^(.+?)-?\d{5,}(-of-\d{5,})?$',
      repl=rf'\g<1>{replacement}',
      string=filepath,
  )
  if num_subs != 1:
    raise RuntimeError(
        f'Should do 1 shard suffix substitution, but did {num_subs}! '
        f'Filepath was {filepath}'
    )
  return new_string


def _filename_template_to_regex(filename_template: str) -> str:
  """Returns the regular expression for the given template.

  Arguments:
    filename_template: the filename template to create a regex for.

  Returns:
    the regular expression for the filename template.

  Raises:
    ValueError: when not all variables in the template were substituted.
  """
  result = filename_template.replace('.', r'\.')
  for var, regex in _VAR_REGEX_MAPPING.items():
    result = result.replace(f'{{{var}}}', regex)
  if re.match(re.compile(r'\{\w+\}'), result):
    raise ValueError(
        'Regex still contains variables '
        f'that have not been substituted: {result}'
    )
  return result


@dataclasses.dataclass()
class ShardedFileTemplate:
  """Template to produce filenames for sharded datasets.

  Attributes:
    data_dir: the directory that contains the files for the shards.
    template: template of the sharded files, e.g.
      '${SPLIT}/data.${FILEFORMAT}-${SHARD_INDEX}'.
    dataset_name: the name of the dataset.
    split: the split of the dataset.
    filetype_suffix: the filetype suffix to denote the type of file. For
      example, `tfrecord`.
  """

  data_dir: epath.Path
  template: str = DEFAULT_FILENAME_TEMPLATE
  dataset_name: Optional[str] = None
  split: Optional[str] = None
  filetype_suffix: Optional[str] = None

  def __post_init__(self):
    self.data_dir = epath.Path(self.data_dir)
    if self.split is not None and not self.split:
      raise ValueError(f'Split must be a non-empty string: {self}')
    if self.split is not None and not any(
        char.isalnum() for char in self.split
    ):
      raise ValueError(
          'Split name should contain at least one alphanumeric character.'
          f' Given split name: {self.split}'
      )
    if self.filetype_suffix is not None and not self.filetype_suffix:
      raise ValueError(f'Filetype suffix must be a non-empty string: {self}')
    if not self.template:
      self.template = DEFAULT_FILENAME_TEMPLATE

  @functools.cached_property
  def regex(self) -> 're.Pattern[str]':
    """Returns the regular expression for this template.

    Can be used to test whether a filename matches to this template.
    """
    return re.compile(_filename_template_to_regex(self.template))

  def parse_filename_info(self, filename: str) -> Optional[FilenameInfo]:
    """Parses the filename using this template.

    Note that when the filename doesn't specify the dataset name, split, or
    filetype suffix, but this template does, then the value in the template will
    be used.

    Arguments:
      filename: the filename that should be parsed.

    Returns:
      the FilenameInfo corresponding to the given file if it could be parsed.
      None otherwise.
    """
    match = self.regex.fullmatch(filename)
    if not match:
      return None
    groupdict = match.groupdict()
    shard_index = groupdict.get('shard_index')
    num_shards = groupdict.get('num_shards')
    return FilenameInfo(
        dataset_name=groupdict.get('dataset_name', self.dataset_name),
        split=groupdict.get('split', self.split),
        filetype_suffix=groupdict.get('filetype_suffix', self.filetype_suffix),
        shard_index=int(shard_index) if shard_index is not None else None,
        num_shards=int(num_shards) if num_shards is not None else None,
        filename_template=self,
    )

  def is_valid(self, filename: str) -> bool:
    """Returns whether the given filename follows this template."""
    filename_info = self.parse_filename_info(filename)
    if filename_info is None:
      return False

    # Even when `dataset_name` is set, it may not be in the template,
    # so also test that `filename_info.dataset_name` is not None`.`
    if (
        self.dataset_name is not None
        and filename_info.dataset_name is not None
        and filename_info.dataset_name != self.dataset_name
    ):
      return False
    if (
        self.split is not None
        and filename_info.split is not None
        and filename_info.split != self.split
    ):
      return False
    if (
        self.filetype_suffix is not None
        and filename_info.filetype_suffix is not None
        and filename_info.filetype_suffix != self.filetype_suffix
    ):
      return False
    return True

  def _default_mappings(self) -> MutableMapping[str, Any]:
    mappings = {}
    if self.split:
      mappings[_VAR_SPLIT] = self.split
    if self.dataset_name:
      mappings[_VAR_DATASET] = self.dataset_name
    if self.filetype_suffix:
      mappings[_VAR_FILEFORMAT] = self.filetype_suffix
    return mappings

  def relative_filepath(
      self,
      *,
      shard_index: int,
      num_shards: Optional[int],
  ) -> str:
    """Returns the path (relative to the data dir) of the shard."""
    mappings = self._default_mappings()

    # Add shard related information that is formatted consistently
    shard_number_template = '{n:0%d}' % _num_digits_needed(num_shards)
    mappings[_VAR_SHARD_INDEX] = shard_number_template.format(n=shard_index)
    if num_shards:
      mappings[_VAR_NUM_SHARDS] = shard_number_template.format(n=num_shards)
      mappings[_VAR_SHARD_X_OF_Y] = (
          f'{mappings[_VAR_SHARD_INDEX]}-of-{mappings[_VAR_NUM_SHARDS]}'
      )
    return self.template.format(**mappings)

  def sharded_filepath(
      self,
      *,
      shard_index: int,
      num_shards: Optional[int],
  ) -> epath.Path:
    """Returns the filename (including full path if `data_dir` is set) for the given shard."""
    return self.data_dir / self.relative_filepath(
        shard_index=shard_index, num_shards=num_shards
    )

  def sharded_filepaths(
      self,
      num_shards: int,
  ) -> List[epath.Path]:
    return [
        self.sharded_filepath(shard_index=i, num_shards=num_shards)
        for i in range(num_shards)
    ]

  def filepath_prefix(
      self,
  ) -> str:
    a_filepath = self.sharded_filepath(shard_index=0, num_shards=1)
    return _replace_shard_suffix(os.fspath(a_filepath), '')

  def sharded_filepaths_pattern(
      self,
      *,
      num_shards: Optional[int] = None,
  ) -> str:
    """Returns a pattern describing all the file paths captured by this template.

    If `num_shards` is given, then it returns
    '/path/dataset_name-split.fileformat@num_shards`.
    If `num_shards` is not given, then it returns
    '/path/dataset_name-split.fileformat*`.

    Args:
      num_shards: optional specification of the number of shards.

    Returns:
      the pattern describing all shards captured by this template.
    """
    a_filepath = self.sharded_filepath(shard_index=0, num_shards=1)
    if num_shards:
      replacement = f'@{num_shards}'
    else:
      replacement = '*'
    return _replace_shard_suffix(os.fspath(a_filepath), replacement)

  def sharded_filenames(self, num_shards: int) -> List[str]:
    return [path.name for path in self.sharded_filepaths(num_shards=num_shards)]

  def replace(self, **kwargs: Any) -> 'ShardedFileTemplate':
    """Returns a copy of the `ShardedFileTemplate` with updated attributes."""
    return dataclasses.replace(self, **kwargs)


def filepattern_for_dataset_split(
    *,
    dataset_name: str,
    split: str,
    data_dir: str,
    filetype_suffix: Optional[str] = None,
    num_shards: Optional[int] = None,
) -> str:
  """Returns the file pattern for the given dataset.

  TODO(tfds): remove this by start using ShardedFileTemplate

  Args:
    dataset_name: Name of the dataset
    split: Name of the requested split
    data_dir: The base folder that contains the dataset.
    filetype_suffix: Optional suffix, e.g. tfrecord
    num_shards: Optional argument. If specified, will return file@num_shards
      notation, otherwise file*.
  """
  template = ShardedFileTemplate(
      data_dir=epath.Path(data_dir),
      dataset_name=dataset_name,
      split=split,
      filetype_suffix=filetype_suffix,
  )
  return os.fspath(template.sharded_filepaths_pattern(num_shards=num_shards))


def filenames_for_dataset_split(
    dataset_name: str,
    split: str,
    num_shards: int,
    filetype_suffix: str,
    data_dir: Optional[epath.PathLike] = None,
) -> List[str]:
  """Returns the list of filenames for the given dataset and split."""
  # TODO(tfds): remove this by start using ShardedFileTemplate
  template = ShardedFileTemplate(
      dataset_name=dataset_name,
      split=split,
      filetype_suffix=filetype_suffix,
      data_dir=epath.Path(data_dir),
  )
  return [
      os.fspath(fp) for fp in template.sharded_filenames(num_shards=num_shards)
  ]


def filepaths_for_dataset_split(
    dataset_name: str,
    split: str,
    num_shards: int,
    data_dir: str,
    filetype_suffix: str,
) -> List[str]:
  """File paths of a given dataset split."""
  # TODO(tfds): remove this by start using ShardedFileTemplate
  template = ShardedFileTemplate(
      dataset_name=dataset_name,
      split=split,
      filetype_suffix=filetype_suffix,
      data_dir=epath.Path(data_dir),
  )
  return [
      os.fspath(fp) for fp in template.sharded_filepaths(num_shards=num_shards)
  ]


def _get_filename_template(
    filename: str, filename_template: Optional[ShardedFileTemplate]
) -> ShardedFileTemplate:
  if filename_template is None:
    return ShardedFileTemplate(data_dir=epath.Path(os.path.dirname(filename)))
  return filename_template


@dataclasses.dataclass(eq=True, frozen=True)
class FilenameInfo:
  """Structure representing a filename.

  Attributes:
    dataset_name: the name of the dataset, e.g. `mnist`.
    split: the split to which this file belongs, e.g. `train` or `test`.
    filetype_suffix: the suffix representing the filetype, e.g. `tfrecord`.
    shard_index: what shard this file is.
    num_shards: if known, the total number of shards.
    filename_template: the template to which this file conforms.
  """

  dataset_name: Optional[str] = None
  split: Optional[str] = None
  filetype_suffix: Optional[str] = None
  shard_index: Optional[int] = None
  num_shards: Optional[int] = None
  filename_template: Optional[ShardedFileTemplate] = None

  def full_filename_template(self):
    template = self.filename_template or ShardedFileTemplate(
        data_dir=epath.Path('')
    )
    return template.replace(
        dataset_name=self.dataset_name,
        split=self.split,
        filetype_suffix=self.filetype_suffix,
    )

  def replace(self, **kwargs: Any) -> 'FilenameInfo':
    """Returns a copy with updated attributes."""
    return dataclasses.replace(self, **kwargs)

  @classmethod
  def from_str(
      cls,
      filename: str,
      filename_template: Optional[ShardedFileTemplate] = None,
  ) -> 'FilenameInfo':
    """Factory to create a `FilenameInfo` from filename."""
    filename_template = _get_filename_template(filename, filename_template)
    # Strip off the directory if the filename contains it.
    filename = os.path.basename(filename)
    filename_info = filename_template.parse_filename_info(filename)
    if filename_info is None:
      raise ValueError(
          f'Could not parse filename {filename} '
          f'with template {filename_template}'
      )
    return filename_info

  @staticmethod
  def is_valid(
      filename: str,
      filename_template: Optional[ShardedFileTemplate] = None,
  ) -> bool:
    """Returns True if the filename follow the given pattern."""
    filename_template = _get_filename_template(filename, filename_template)
    return filename_template.is_valid(filename)

  def __str__(self) -> str:
    return self.full_filename_template().relative_filepath(
        shard_index=self.shard_index, num_shards=self.num_shards
    )



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Beam utils."""

import functools
from typing import Any

from tensorflow_datasets.core import dataset_builder
from tensorflow_datasets.core import lazy_imports_lib
from tensorflow_datasets.core import naming
from tensorflow_datasets.core.utils import shard_utils

__all__ = [
    'inc_counter',
    'ReadFromTFDS',
]


@lazy_imports_lib.beam_ptransform_fn
def ReadFromTFDS(  # pylint: disable=invalid-name
    pipeline,
    builder: dataset_builder.DatasetBuilder,
    split: str,
    workers_per_shard: int = 1,
    **as_dataset_kwargs: Any,
):
  """Creates a beam pipeline yielding TFDS examples.

  Each dataset shard will be processed in parallel.

  Usage:

  ```python
  builder = tfds.builder('my_dataset')

  _ = (
      pipeline
      | tfds.beam.ReadFromTFDS(builder, split='train')
      | beam.Map(tfds.as_numpy)
      | ...
  )
  ```

  Use `tfds.as_numpy` to convert each examples from `tf.Tensor` to numpy.

  The split argument can make use of subsplits, eg 'train[:100]', only when the
  batch_size=None (in as_dataset_kwargs). Note: the order of the images will be
  different than when tfds.load(split='train[:100]') is used, but the same
  examples will be used.

  Args:
    pipeline: beam pipeline (automatically set)
    builder: Dataset builder to load
    split: Split name to load (e.g. `train+test`, `train`)
    workers_per_shard: number of workers that should read a shard in parallel.
      The shard will be split in this many parts. Note that workers cannot skip
      to a specific row in a tfrecord file, so they need to read the file up
      until that point without using that data.
    **as_dataset_kwargs: Arguments forwarded to `builder.as_dataset`.

  Returns:
    The PCollection containing the TFDS examples.
  """
  beam = lazy_imports_lib.lazy_imports.apache_beam

  if not builder.info.splits.total_num_examples:
    raise ValueError(
        f'No examples found in {builder.name!r} in data dir {builder.data_dir}.'
        ' Was the dataset generated?'
    )

  def load_shard(file_instruction: shard_utils.FileInstruction):  # pylint: disable=invalid-name
    """Loads a single shard."""
    file_info = naming.FilenameInfo.from_str(file_instruction.filename)
    ds = builder.as_dataset(
        split=f'{file_info.split}[{file_info.shard_index}shard]',
        **as_dataset_kwargs,
    )

    # Subsplit is supported by adding ds.skip and ds.take, which will only work
    # if `batch_size=None. An alternativ would be to convert file_instruction
    # to absolute instructions (and check that
    # `splits[abs_split].file_instructions == [file_instruction]` ).

    if file_instruction.skip > 0 or not file_instruction.takes_all:
      batch_size = as_dataset_kwargs.get('batch_size')
      if batch_size is not None:
        raise NotImplementedError(
            'ReadFromTFDS supports skip and take (used '
            f'in split={split!r}) only when batch_size='
            f'None. Got batch_size={batch_size!r}.'
        )

    if file_instruction.skip > 0:
      ds = ds.skip(file_instruction.skip)
    if not file_instruction.takes_all:
      ds = ds.take(file_instruction.take)
    inc_counter(
        name='LoadedFileInstructions', value=1, namespace='ReadFromTFDS'
    )
    return ds

  file_instructions = builder.info.splits[split].file_instructions
  inc_counter(
      name='FileInstructions',
      value=len(file_instructions),
      namespace='ReadFromTFDS',
  )
  if workers_per_shard > 1:
    expanded_file_instructions = []
    for file_instruction in file_instructions:
      expanded_file_instructions.extend(
          shard_utils.split_file_instruction(
              file_instruction=file_instruction, num_splits=workers_per_shard
          )
      )
    file_instructions = expanded_file_instructions
    inc_counter(
        name='ExpandedFileInstructions',
        value=len(file_instructions),
        namespace='ReadFromTFDS',
    )
  return pipeline | beam.Create(file_instructions) | beam.FlatMap(load_shard)


@functools.lru_cache(None)
def _get_metrics_counter(namespace: str, name: str):
  beam = lazy_imports_lib.lazy_imports.apache_beam
  return beam.metrics.Metrics.counter(namespace, name)


def inc_counter(name: str, value: int = 1, *, namespace: str = 'tfds') -> None:
  """Increment a `beam.metrics.Metrics.counter` without boilerplate.

  ```python
  tfds.beam.inc_counter('my_metric')
  ```

  Is a small alias for:

  ```python
  beam.metrics.Metrics.counter(namespace, name).inc(value)
  ```

  Args:
    name: The metrics name
    value: Increment (default to 1)
    namespace: Metrics namespace (default to `tfds`)
  """
  _get_metrics_counter(namespace, name).inc(value)



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Hillstrom dataset."""

from __future__ import annotations

import csv

import numpy as np
import tensorflow_datasets.public_api as tfds

_DESCRIPTION = """
This dataset contains 64,000 customers who last purchased within twelve months. The customers were involved in an e-mail test.

1. 1/3 were randomly chosen to receive an e-mail campaign featuring Mens merchandise.
2. 1/3 were randomly chosen to receive an e-mail campaign featuring Womens merchandise.
3. 1/3 were randomly chosen to not receive an e-mail campaign.

During a period of two weeks following the e-mail campaign, results were tracked.
The task is to tell the world if the Mens or Womens e-mail campaign was successful.
"""

_CITATION = """
@article{entryhillstrom,
  title={Hillstrom’s MineThatData Email Analytics Challenge},
  author={ENTRY, WINNING}
}
"""


class Hillstrom(tfds.core.GeneratorBasedBuilder):
  """DatasetBuilder for dataset."""

  VERSION = tfds.core.Version('1.0.0')
  RELEASE_NOTES = {
      '1.0.0': 'Initial release.',
  }

  def _info(self) -> tfds.core.DatasetInfo:
    """Returns the dataset metadata."""
    return tfds.core.DatasetInfo(
        builder=self,
        description=_DESCRIPTION,
        features=tfds.features.FeaturesDict({
            # These are the features of your dataset like images, labels ...
            'history': np.float32,
            'zip_code': tfds.features.Text(),
            'segment': tfds.features.Text(),
            'recency': np.int64,
            'history_segment': tfds.features.Text(),
            'mens': np.int64,
            'womens': np.int64,
            'newbie': np.int64,
            'channel': tfds.features.Text(),
            'visit': np.int64,
            'conversion': np.int64,
            'spend': np.float32,
        }),
        # If there's a common (input, target) tuple from the
        # features, specify them here. They'll be used if
        # `as_supervised=True` in `builder.as_dataset`.
        supervised_keys=(
            {
                'history': 'history',
                'zip_code': 'zip_code',
                'segment': 'segment',
                'recency': 'recency',
                'mens': 'mens',
                'womens': 'womens',
                'newbie': 'newbie',
                'channel': 'channel',
            },
            'visit',
        ),
        homepage='https://blog.minethatdata.com/2008/03/minethatdata-e-mail-analytics-and-data.html',
        citation=_CITATION,
    )

  def _split_generators(self, dl_manager: tfds.download.DownloadManager):
    """Returns SplitGenerators."""
    path = dl_manager.download_and_extract(
        'http://www.minethatdata.com/Kevin_Hillstrom_MineThatData_E-MailAnalytics_DataMiningChallenge_2008.03.20.csv'
    )

    return {
        'train': self._generate_examples(path),
    }

  def _generate_examples(self, path):
    """Yields examples."""
    with path.open() as f:
      index = 0
      for row in csv.DictReader(f):
        # And yield (key, feature_dict)
        yield index, {
            'recency': row['recency'],
            'history_segment': row['history_segment'],
            'mens': row['mens'],
            'womens': row['womens'],
            'newbie': row['newbie'],
            'channel': row['channel'],
            'segment': row['segment'],
            'visit': row['visit'],
            'conversion': row['conversion'],
            'spend': row['spend'],
            'history': row['history'],
            'zip_code': row['zip_code'],
        }
        index += 1



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""wiki_auto dataset."""

import sys
import tensorflow_datasets.public_api as tfds
from tensorflow_datasets.text_simplification.wiki_auto import wiki_auto

sys.path.insert(0, '.')


class WikiAutoTestMANUAL(tfds.testing.DatasetBuilderTestCase):
  """Tests for wiki_auto dataset for manual configuration."""

  BUILDER_CONFIG_NAMES_TO_TEST = ['manual']
  DATASET_CLASS = wiki_auto.WikiAuto

  DL_EXTRACT_RESULT = {
      'train': 'manual/1.0.0/train.tsv',
      'dev': 'manual/1.0.0/dev.tsv',
      'test': 'manual/1.0.0/test.tsv',
  }
  SPLITS = {
      'train': 5,
      'dev': 5,
      'test': 5,
  }


class WikiAutoTestAUTO(tfds.testing.DatasetBuilderTestCase):
  """Tests for wiki_auto dataset for auto configuration."""

  BUILDER_CONFIG_NAMES_TO_TEST = ['auto']
  DATASET_CLASS = wiki_auto.WikiAuto

  DL_EXTRACT_RESULT = {
      'part1': 'auto/1.0.0/wiki-auto-part-1-data.json',
      'part2': 'auto/1.0.0/wiki-auto-part-2-data.json',
  }
  SPLITS = {
      'part1': 1,
      'part2': 1,
  }


class WikiAutoTestAUTOACL(tfds.testing.DatasetBuilderTestCase):
  """Tests for wiki_auto dataset for auto_acl configuration."""

  BUILDER_CONFIG_NAMES_TO_TEST = ['auto_acl']
  DATASET_CLASS = wiki_auto.WikiAuto

  DL_EXTRACT_RESULT = {
      'normal': 'auto_acl/1.0.0/train.dst',
      'simple': 'auto_acl/1.0.0/train.src',
  }
  SPLITS = {'full': 5}


class WikiAutoTestAUTOACLNOSPLIT(tfds.testing.DatasetBuilderTestCase):
  """Tests for wiki_auto dataset for auto_acl configuration."""

  BUILDER_CONFIG_NAMES_TO_TEST = ['auto_full_no_split']
  DATASET_CLASS = wiki_auto.WikiAuto

  DL_EXTRACT_RESULT = {
      'normal': 'auto_full_no_split/1.0.0/train.dst',
      'simple': 'auto_full_no_split/1.0.0/train.src',
  }
  SPLITS = {'full': 5}


class WikiAutoTestAUTOACLWITHSPLIT(tfds.testing.DatasetBuilderTestCase):
  """Tests for wiki_auto dataset for auto_acl configuration."""

  BUILDER_CONFIG_NAMES_TO_TEST = ['auto_full_with_split']
  DATASET_CLASS = wiki_auto.WikiAuto

  DL_EXTRACT_RESULT = {
      'normal': 'auto_full_with_split/1.0.0/train.dst',
      'simple': 'auto_full_with_split/1.0.0/train.src',
  }
  SPLITS = {'full': 5}


if __name__ == '__main__':
  tfds.testing.test_main()



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for tensorflow_datasets.core.community.register_path."""

import contextlib
import datetime
import json
import os
import pathlib
import sys
import tempfile
import textwrap
from typing import Iterator
from unittest import mock

from etils import epath
import pytest

from tensorflow_datasets.core import dataset_builder
from tensorflow_datasets.core import naming
from tensorflow_datasets.core import registered
from tensorflow_datasets.core import utils
from tensorflow_datasets.core.community import cache
from tensorflow_datasets.core.community import dataset_sources
from tensorflow_datasets.core.community import register_package


@contextlib.contextmanager
def mock_cache_path(new_cache_dir: epath.PathLike) -> Iterator[None]:
  """Mock which overwrite the cache path."""
  new_dir = epath.Path(new_cache_dir)

  # Use `__wrapped__` to access the original function wrapped inside
  # `functools.lru_cache`
  new_cache_path = utils.memoize()(cache.cache_path.__wrapped__)
  new_module_path = utils.memoize()(cache.module_path.__wrapped__)
  with mock.patch.object(
      cache, '_default_cache_dir', return_value=new_dir
  ), mock.patch.object(cache, 'cache_path', new_cache_path), mock.patch.object(
      cache, 'module_path', new_module_path
  ):
    yield


@pytest.fixture(scope='module')
def dummy_register():
  """Dummy register."""

  with tempfile.TemporaryDirectory() as tmp_path:
    tmp_path = pathlib.Path(tmp_path)

    source_path = utils.tfds_path() / 'testing/dummy_dataset/dummy_dataset.py'

    # Single-file dataset package (without checksums)
    src_single = dataset_sources.DatasetSource.from_json(os.fspath(source_path))

    # Multi-file dataset package (with checksums)
    src_multi = dataset_sources.DatasetSource.from_json({
        'root_path': os.fspath(source_path.parent),
        'filenames': ['checksums.tsv', 'dummy_dataset.py'],
    })
    src_multi_json = json.dumps(src_multi.to_json())  # `dict` -> `str`

    # Create the remote index content
    # Note the absence of `"` for the `src_multi_json` as it is parsed as `dict`
    content = textwrap.dedent(
        f"""\
        {{"name": "kaggle:dummy_dataset", "source": "{src_single.to_json()}"}}
        {{"name": "kaggle:ds1", "source": "{src_single.to_json()}"}}
        {{"name": "mlds:dummy_dataset", "source": {src_multi_json}}}
        """
    )
    dummy_path = tmp_path / 'dummy-community-datasets.toml'
    dummy_path.write_text(content)

    with mock_cache_path(tmp_path / 'cache'):
      yield register_package.PackageRegister(path=dummy_path)


def test_list_dataset_references(dummy_register):  # pylint: disable=redefined-outer-name
  assert sorted(dummy_register.list_dataset_references()) == [
      naming.DatasetReference(dataset_name='ds1', namespace='kaggle'),
      naming.DatasetReference(dataset_name='dummy_dataset', namespace='kaggle'),
      naming.DatasetReference(dataset_name='dummy_dataset', namespace='mlds'),
  ]


def test_builder_cls(dummy_register):  # pylint: disable=redefined-outer-name
  # The dataset will be installed in the cache
  installed_path = cache.cache_path()
  installed_path /= 'modules/tfds_community/kaggle/dummy_dataset'
  assert not installed_path.exists()

  ds_name = naming.DatasetName('kaggle:dummy_dataset')
  builder_cls = dummy_register.builder_cls(ds_name)
  assert builder_cls.name == 'dummy_dataset'

  clshash = 'd183643ab2f6544350b6871f6cc2b0fba07297d8d30f4c04649a7717e44e531b'
  assert installed_path / f'{clshash}/dummy_dataset.py' == builder_cls.code_path
  assert 'kaggle' in builder_cls.code_path.parts
  assert issubclass(builder_cls, dataset_builder.DatasetBuilder)
  assert not builder_cls.url_infos  # No checksums installed with the package

  # Dataset installed in the cache
  # Filename should be deterministic
  assert list(sorted(installed_path.iterdir())) == [installed_path / clshash]

  # Reusing the dataset should re-use the cache
  with mock.patch.object(
      register_package,
      '_download_and_cache',
      side_effect=ValueError('Dataset should have been cached already'),
  ):
    ds_name = naming.DatasetName('kaggle:dummy_dataset')
    builder_cls2 = dummy_register.builder_cls(ds_name)
  assert builder_cls is builder_cls2

  # Datasets from different namespace can have the same name
  ds_name = naming.DatasetName('mlds:dummy_dataset')
  builder_cls = dummy_register.builder_cls(ds_name)
  assert 'mlds' in builder_cls.code_path.parts
  assert issubclass(builder_cls, dataset_builder.DatasetBuilder)
  # Checksums have been correctly installed
  assert 'http://dummy.org/data.txt' in builder_cls.url_infos

  with pytest.raises(registered.DatasetNotFoundError):
    dummy_register.builder(naming.DatasetName('other:ds0'))


def test_register_path_list_builders(dummy_register):  # pylint: disable=redefined-outer-name
  assert dummy_register.list_builders() == [
      'kaggle:ds1',
      'kaggle:dummy_dataset',
      'mlds:dummy_dataset',
  ]


def test_dataset_package():
  """Exports/imports operation should be identity."""
  pkg = register_package.DatasetPackage(
      name=naming.DatasetName('ns:ds'),
      source=dataset_sources.DatasetSource.from_json(
          'github://<owner>/<name>/tree/<branch>/my_ds/ds.py',
      ),
  )
  assert register_package.DatasetPackage.from_json(pkg.to_json()) == pkg

  pkg2 = register_package._InstalledPackage(
      package=pkg,
      instalation_date=datetime.datetime.now(),
      hash='asdajhdadsadsad',
  )
  assert register_package._InstalledPackage.from_json(pkg2.to_json()) == pkg2


def test_mock_cache_path(tmp_path: pathlib.Path):
  with mock_cache_path(tmp_path):
    assert os.fspath(tmp_path) not in sys.path
    assert cache.cache_path() == tmp_path
    assert cache.module_path() == tmp_path / 'modules'
    assert os.fspath(tmp_path / 'modules') in sys.path



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Bounding boxes feature."""

from __future__ import annotations

import collections
from typing import Any, Union

import numpy as np
from tensorflow_datasets.core import lazy_imports_lib
from tensorflow_datasets.core import utils
from tensorflow_datasets.core.features import feature as feature_lib
from tensorflow_datasets.core.features import image_feature
from tensorflow_datasets.core.features import tensor_feature
from tensorflow_datasets.core.proto import feature_pb2
from tensorflow_datasets.core.utils import type_utils

Json = type_utils.Json

BBox = collections.namedtuple('BBox', 'ymin, xmin, ymax, xmax')

PilImage = Any


class BBoxFeature(tensor_feature.Tensor):
  """`FeatureConnector` for a normalized bounding box.

  Note: If you have multiple bounding boxes, you may want to wrap the feature
  inside a `tfds.features.Sequence`.

  Input:
    * `tfds.features.BBox` tuple or 4-dim `np.ndarray`.

  Output:
    bbox: tf.Tensor of type `tf.float32` and shape `[4,]` which contains the
      normalized coordinates of the bounding box `[ymin, xmin, ymax, xmax]`

  Example:
    * In the DatasetInfo object:

    ```
    features=features.FeatureDict({
        'bbox': tfds.features.BBox(),
    })
    ```

    * During generation:

    ```
    yield {
        'input': tfds.features.BBox(ymin=0.3, xmin=0.8, ymax=0.5, xmax=1.0),
    }
    ```
  """

  def __init__(
      self,
      *,
      doc: feature_lib.DocArg = None,
  ):
    super(BBoxFeature, self).__init__(shape=(4,), dtype=np.float32, doc=doc)

  def encode_example(self, bbox: Union[BBox, np.ndarray]):
    """See base class for details."""

    if isinstance(bbox, np.ndarray):
      if bbox.shape != (4,):
        raise ValueError(
            'array representing BBox should have exactly 4 floats. '
            f'Instead, it has {bbox.shape}.'
        )
      bbox = bbox.astype(np.float64)
      bbox = BBox(ymin=bbox[0], xmin=bbox[1], ymax=bbox[2], xmax=bbox[3])

    # Validate the coordinates
    for coordinate in bbox:
      if not isinstance(coordinate, float):
        raise ValueError(
            'BBox coordinates should be float. Got {}.'.format(bbox)
        )
      if not 0.0 <= coordinate <= 1.0:
        raise ValueError(
            'BBox coordinates should be between 0 and 1. Got {}.'.format(bbox)
        )
      if bbox.xmax < bbox.xmin or bbox.ymax < bbox.ymin:
        raise ValueError(
            'BBox coordinates should have min <= max. Got {}.'.format(bbox)
        )

    return super(BBoxFeature, self).encode_example(
        [bbox.ymin, bbox.xmin, bbox.ymax, bbox.xmax]
    )

  def repr_html(self, ex: np.ndarray) -> str:
    """Returns the HTML str representation of an Image with BBoxes."""
    ex = np.expand_dims(ex, axis=0)  # Expand single bounding box to batch.
    return _repr_html(ex)

  def repr_html_batch(self, ex: np.ndarray) -> str:
    """Returns the HTML str representation of an Image with BBoxes (Sequence)."""
    return _repr_html(ex)

  @classmethod
  def from_json_content(
      cls, value: Union[Json, feature_pb2.BoundingBoxFeature]
  ) -> 'BBoxFeature':
    del value  # Unused
    return cls()

  def to_json_content(self) -> feature_pb2.BoundingBoxFeature:  # pytype: disable=signature-mismatch  # overriding-return-type-checks
    return feature_pb2.BoundingBoxFeature(
        shape=feature_lib.to_shape_proto(self._shape),
        dtype=feature_lib.dtype_to_str(self._dtype),
    )


def _repr_html(ex: np.ndarray) -> str:
  """Returns the HTML str representation of an Image with BBoxes."""
  img = _build_thumbnail_with_bbox(ex)
  img_str = utils.get_base64(lambda buff: img.save(buff, format='PNG'))
  return f'<img src="data:image/png;base64,{img_str}" alt="Img" />'


def _build_thumbnail_with_bbox(ex: np.ndarray) -> PilImage:
  """Returns blank image with Bboxes drawn on it."""
  PIL_Image = lazy_imports_lib.lazy_imports.PIL_Image  # pylint: disable=invalid-name
  PIL_ImageDraw = lazy_imports_lib.lazy_imports.PIL_ImageDraw  # pylint: disable=invalid-name

  shape = (image_feature.THUMBNAIL_SIZE, image_feature.THUMBNAIL_SIZE)
  blank_img = PIL_Image.new('RGB', shape, (255, 255, 255))
  draw = PIL_ImageDraw.Draw(blank_img)
  rs = np.random.RandomState(97531)  # freeze random state

  for i in range(ex.shape[0]):
    # Rescale coordinates to match size of blank_image
    ymin, xmin, ymax, xmax = ex[i, :] * image_feature.THUMBNAIL_SIZE
    # Generate random rgb values for Bbox ouline
    r, g, b = list(rs.randint(0, 256, size=3))
    draw.rectangle(((xmin, ymin), (xmax, ymax)), outline=(r, g, b))
  return blank_img



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""wiki_table_questions dataset."""

import csv
from unittest import mock

from etils import epath
import tensorflow_datasets.public_api as tfds
from tensorflow_datasets.structured.wiki_table_questions import wiki_table_questions


class WikiTableQuestionsTest(tfds.testing.DatasetBuilderTestCase):
  """Tests for wiki_table_questions dataset."""

  DATASET_CLASS = wiki_table_questions.WikiTableQuestions
  SPLITS = {
      'train': 2,
      'split-1-train': 1,
      'split-1-dev': 1,
      'split-2-train': 1,
      'split-2-dev': 1,
      'split-3-train': 1,
      'split-3-dev': 1,
      'test': 2,
  }

  OVERLAPPING_SPLITS = ['split-{}-train'.format(i) for i in range(1, 4)] + [
      'split-{}-dev'.format(i) for i in range(1, 4)
  ]

  def test_generate_examples(self):
    examples_file = [
        {
            'id': 'nu-0',
            'utterance': (
                'which country had the most cyclists finish within the top 5?'
            ),
            'context': 'csv/203-csv/733.csv',
            'targetValue': 'Italy',
        },
        {
            'id': 'nu-165',
            'utterance': 'who was the first cyclist to finish?',
            'context': 'csv/203-csv/733.csv',
            'targetValue': 'Alejandro Valverde',
        },
    ]
    csv_203_csv_733 = [
        {
            'Rank': '1',
            'Cyclist': 'Alejandro Valverde (ESP)',
            'Team': "Caisse d'Epargne",
            'Time': '5h 29\' 10"',
            'UCI ProTour\nPoints': '40',
        },
        {
            'Rank': '2',
            'Cyclist': 'Davide Rebellin (ITA)',
            'Team': 'Gerolsteiner',
            'Time': 's.t.',
            'UCI ProTour\nPoints': '25',
        },
        {
            'Rank': '3',
            'Cyclist': 'Paolo Bettini (ITA)',
            'Team': 'Quick Step',
            'Time': 's.t.',
            'UCI ProTour\nPoints': '20',
        },
    ]
    table_formated = [
        {
            'column_header': 'Rank',
            'row_number': 1,
            'content': '1',
        },
        {
            'column_header': 'Cyclist',
            'row_number': 1,
            'content': 'Alejandro Valverde (ESP)',
        },
        {
            'column_header': 'Team',
            'row_number': 1,
            'content': """Caisse d'Epargne""",
        },
        {
            'column_header': 'Time',
            'row_number': 1,
            'content': '''5h 29' 10"''',
        },
        {
            'column_header': 'UCI ProTour\nPoints',
            'row_number': 1,
            'content': '40',
        },
        {
            'column_header': 'Rank',
            'row_number': 2,
            'content': '2',
        },
        {
            'column_header': 'Cyclist',
            'row_number': 2,
            'content': 'Davide Rebellin (ITA)',
        },
        {
            'column_header': 'Team',
            'row_number': 2,
            'content': 'Gerolsteiner',
        },
        {
            'column_header': 'Time',
            'row_number': 2,
            'content': 's.t.',
        },
        {
            'column_header': 'UCI ProTour\nPoints',
            'row_number': 2,
            'content': '25',
        },
        {
            'column_header': 'Rank',
            'row_number': 3,
            'content': '3',
        },
        {
            'column_header': 'Cyclist',
            'row_number': 3,
            'content': 'Paolo Bettini (ITA)',
        },
        {
            'column_header': 'Team',
            'row_number': 3,
            'content': 'Quick Step',
        },
        {
            'column_header': 'Time',
            'row_number': 3,
            'content': 's.t.',
        },
        {
            'column_header': 'UCI ProTour\nPoints',
            'row_number': 3,
            'content': '20',
        },
    ]
    expected_examples = [
        {
            'input_text': {
                'table': table_formated,
                'context': (
                    'which country had the most cyclists finish within the'
                    ' top 5?'
                ),
            },
            'target_text': 'Italy',
        },
        {
            'input_text': {
                'table': table_formated,
                'context': 'who was the first cyclist to finish?',
            },
            'target_text': 'Alejandro Valverde',
        },
    ]
    with mock.patch.object(
        csv,
        'DictReader',
        side_effect=iter([examples_file, csv_203_csv_733, csv_203_csv_733]),
    ):
      mock_open = mock.MagicMock()
      with epath.testing.mock_epath(open=mock_open):
        dataset = wiki_table_questions.WikiTableQuestions()
        examples = list(dataset._generate_examples('', ''))
        for i, (_, example) in enumerate(examples):
          self.assertCountEqual(example, expected_examples[i])
        assert len(examples) == len(expected_examples)


if __name__ == '__main__':
  tfds.testing.test_main()



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Even split utils."""

import dataclasses
import functools
import operator
from typing import List, Optional

from absl import logging
from tensorflow_datasets.core import lazy_imports_lib
from tensorflow_datasets.core import splits as splits_lib


@dataclasses.dataclass(frozen=True)
class _EvenSplit(splits_lib.AbstractSplit):
  """Split matching a subsplit of the given split."""

  split: splits_lib.SplitArg
  index: int
  count: int
  drop_remainder: bool

  def to_absolute(self, split_infos: splits_lib.SplitDict):
    # Extract the absolute instructions
    # One absolute instruction is created per `+`, so `train[:54%]+train[60%:]`
    # will create 2 absolute instructions.
    read_instruction = splits_lib.AbstractSplit.from_spec(self.split)
    absolute_instructions = read_instruction.to_absolute(split_infos)

    # Create the subsplit
    read_instructions_for_index = [
        self._absolute_to_read_instruction_for_index(abs_inst, split_infos)
        for abs_inst in absolute_instructions
    ]
    instuctions = functools.reduce(operator.add, read_instructions_for_index)
    return instuctions.to_absolute(split_infos)

  def _absolute_to_read_instruction_for_index(
      self,
      abs_inst,
      split_infos: splits_lib.SplitDict,
  ) -> splits_lib.ReadInstruction:
    start = abs_inst.from_ or 0
    if abs_inst.to is None:  # Note: `abs_inst.to == 0` is valid
      end = split_infos[abs_inst.splitname].num_examples
    else:
      end = abs_inst.to

    assert end >= start, f'start={start}, end={end}'
    num_examples = end - start

    examples_per_split = num_examples // self.count
    split_start = start + examples_per_split * self.index
    split_end = start + examples_per_split * (self.index + 1)

    # Handle remaining examples.
    num_unused_examples = num_examples % self.count
    assert num_unused_examples >= 0, num_unused_examples
    assert num_unused_examples < self.count, num_unused_examples
    if num_unused_examples > 0:
      if self.drop_remainder:
        logging.warning(
            'Dropping %d examples of %d examples (host count: %d).',
            num_unused_examples,
            num_examples,
            self.count,
        )
      else:
        split_start += min(self.index, num_unused_examples)
        split_end += min(self.index + 1, num_unused_examples)

    return splits_lib.ReadInstruction(
        abs_inst.splitname,
        from_=split_start,
        to=split_end,
        unit='abs',
    )


def even_splits(
    split: str,
    n: int,
    *,
    drop_remainder: bool = False,
) -> List[splits_lib.SplitArg]:
  """Generates a list of non-overlapping sub-splits of same size.

  Example:

  ```python
  split0, split1, split2 = tfds.even_splits('train', n=3, drop_remainder=True)

  # Load 1/3 of the train split.
  ds = tfds.load('my_dataset', split=split0)
  ```

  `tfds.even_splits` supports arbitrary
  [sub-splits](https://www.tensorflow.org/datasets/splits) inputs, including
  other `tfds.even_splits` outputs.

  Args:
    split: Split (e.g. 'train', 'test[75%:]',...)
    n: Number of sub-splits to create
    drop_remainder: Drop examples if the number of examples in the datasets is
      not evenly divisible by `n`. If `False`, examples are distributed evenly
      across subsplits, starting by the first. For example, if there is 11
      examples with `n=3`, splits will contain `[4, 4, 3]` examples
      respectivelly.

  Returns:
    The list of subsplits. Those splits can be combined together (with
      `+`) or with other subsplits (e.g. `split + 'test[75%:]'`).
  """
  return [
      _EvenSplit(split=split, index=i, count=n, drop_remainder=drop_remainder)
      for i in range(n)
  ]


def split_for_jax_process(
    split: str,
    *,
    process_index: Optional[int] = None,
    process_count: Optional[int] = None,
    drop_remainder: bool = False,
) -> splits_lib.SplitArg:
  """Returns the subsplit of the data for the process.

  In distributed setting, all process/hosts should get a non-overlapping,
  equally sized slice of the entire data. This function takes as input a split
  and extracts the slice for the current process index.

  Usage:

  ```python
  tfds.load(..., split=tfds.split_for_jax_process('train'))
  ```

  This funtion is an alias for:

  ```python
  tfds.even_splits(split, n=jax.process_count())[jax.process_index()]
  ```

  By default, if examples can't be evenly distributed across processes, you can
  drop extra examples with `drop_remainder=True`.

  Args:
    split: Split to distribute across host (e.g. `train[75%:]`,
      `train[:800]+validation[:100]`).
    process_index: Process index in `[0, count)`. Defaults to
      `jax.process_index()`.
    process_count: Number of processes. Defaults to `jax.process_count()`.
    drop_remainder: Drop examples if the number of examples in the datasets is
      not evenly divisible by `n`. If `False`, examples are distributed evenly
      across subsplits, starting by the first. For example, if there is 11
      examples with `n=3`, splits will contain `[4, 4, 3]` examples
      respectivelly.

  Returns:
    subsplit: The sub-split of the given `split` for the current
      `process_index`.
  """
  if process_index is None:
    process_index = lazy_imports_lib.lazy_imports.jax.process_index()
  if process_count is None:
    process_count = lazy_imports_lib.lazy_imports.jax.process_count()
  return even_splits(
      split,
      n=process_count,
      drop_remainder=drop_remainder,
  )[process_index]



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Locomotion dataset."""

import numpy as np
import tensorflow_datasets.public_api as tfds
from tensorflow_datasets.rlds import rlds_base


_DESCRIPTION = """
The datasets were created with a SAC agent trained on the environment reward of
MuJoCo locomotion tasks. These datasets are used in
[What Matters for Adversarial Imitation Learning? Orsini et al. 2021](https://arxiv.org/pdf/2106.00672.pdf).

The datasets follow the [RLDS format](https://github.com/google-research/rlds)
to represent steps and episodes.s
"""

_CITATION = """
  @article{orsini2021matters,
    title={What Matters for Adversarial Imitation Learning?},
    author={Orsini, Manu and Raichuk, Anton and Hussenot, L{\'e}onard and Vincent, Damien and Dadashi, Robert and Girgin, Sertan and Geist, Matthieu and Bachem, Olivier and Pietquin, Olivier and Andrychowicz, Marcin},
    journal={International Conference in Machine Learning},
    year={2021}
  }
"""

_HOMEPAGE = 'https://github.com/google-research/rlds'


class Locomotion(tfds.core.GeneratorBasedBuilder):
  """DatasetBuilder for locomotion dataset."""

  VERSION = tfds.core.Version('1.0.0')
  RELEASE_NOTES = {
      '1.0.0': 'Initial release.',
  }

  _DATA_PATHS = {
      'ant_sac_1M_single_policy_stochastic': 'https://storage.googleapis.com/rlds_external_data_release/ant_sac_1M_single_policy_stochastic.tar.gz',
      'walker2d_sac_1M_single_policy_stochastic': 'https://storage.googleapis.com/rlds_external_data_release/walker2d_sac_1M_single_policy_stochastic.tar.gz',
      'humanoid_sac_15M_single_policy_stochastic': 'https://storage.googleapis.com/rlds_external_data_release/humanoid_sac_15M_single_policy_stochastic.tar.gz',
      'hopper_sac_1M_single_policy_stochastic': 'https://storage.googleapis.com/rlds_external_data_release/hopper_sac_1M_single_policy_stochastic.tar.gz',
      'halfcheetah_sac_1M_single_policy_stochastic': 'https://storage.googleapis.com/rlds_external_data_release/halfcheetah_sac_1M_single_policy_stochastic.tar.gz',
  }

  BUILDER_CONFIGS = [
      rlds_base.DatasetConfig(
          name='ant_sac_1M_single_policy_stochastic',
          observation_info=tfds.features.Tensor(shape=(111,), dtype=np.float32),
          action_info=tfds.features.Tensor(shape=(8,), dtype=np.float32),
          reward_info=np.float32,
          discount_info=np.float32,
          citation=_CITATION,
          homepage=_HOMEPAGE,
          overall_description=_DESCRIPTION,
          description=(
              'Dataset generated by a SAC agent trained for 1M steps for Ant.'
          ),
          supervised_keys=None,  # pytype: disable=wrong-arg-types  # gen-stub-imports
      ),
      rlds_base.DatasetConfig(
          name='hopper_sac_1M_single_policy_stochastic',
          observation_info=tfds.features.Tensor(shape=(11,), dtype=np.float32),
          action_info=tfds.features.Tensor(shape=(3,), dtype=np.float32),
          reward_info=np.float32,
          discount_info=np.float32,
          citation=_CITATION,
          homepage=_HOMEPAGE,
          overall_description=_DESCRIPTION,
          description=(
              'Dataset generated by a SAC agent trained for 1M steps for'
              ' Hopper.'
          ),
          supervised_keys=None,  # pytype: disable=wrong-arg-types  # gen-stub-imports
      ),
      rlds_base.DatasetConfig(
          name='halfcheetah_sac_1M_single_policy_stochastic',
          observation_info=tfds.features.Tensor(shape=(17,), dtype=np.float32),
          action_info=tfds.features.Tensor(shape=(6,), dtype=np.float32),
          reward_info=np.float32,
          discount_info=np.float32,
          citation=_CITATION,
          homepage=_HOMEPAGE,
          overall_description=_DESCRIPTION,
          description=(
              'Dataset generated by a SAC agent trained for 1M steps for'
              ' HalfCheetah.'
          ),
          supervised_keys=None,  # pytype: disable=wrong-arg-types  # gen-stub-imports
      ),
      rlds_base.DatasetConfig(
          name='walker2d_sac_1M_single_policy_stochastic',
          observation_info=tfds.features.Tensor(shape=(17,), dtype=np.float32),
          action_info=tfds.features.Tensor(shape=(6,), dtype=np.float32),
          reward_info=np.float32,
          discount_info=np.float32,
          citation=_CITATION,
          homepage=_HOMEPAGE,
          overall_description=_DESCRIPTION,
          description=(
              'Dataset generated by a SAC agent trained for 1M steps for'
              ' Walker2d.'
          ),
          supervised_keys=None,  # pytype: disable=wrong-arg-types  # gen-stub-imports
      ),
      rlds_base.DatasetConfig(
          name='humanoid_sac_15M_single_policy_stochastic',
          observation_info=tfds.features.Tensor(shape=(376,), dtype=np.float32),
          action_info=tfds.features.Tensor(shape=(17,), dtype=np.float32),
          reward_info=np.float32,
          discount_info=np.float32,
          citation=_CITATION,
          homepage=_HOMEPAGE,
          overall_description=_DESCRIPTION,
          description=(
              'Dataset generated by a SAC agent trained for 15M steps for'
              ' Humanoid.'
          ),
          supervised_keys=None,  # pytype: disable=wrong-arg-types  # gen-stub-imports
      ),
  ]


  def _info(self) -> tfds.core.DatasetInfo:
    """Returns the dataset metadata."""
    return rlds_base.build_info(self.builder_config, self)

  def _split_generators(self, dl_manager: tfds.download.DownloadManager):
    """Returns SplitGenerators."""
    path = dl_manager.download_and_extract(
        {
            'file_path': self._DATA_PATHS[self.builder_config.name],
        }
    )
    return {
        'train': self._generate_examples(path),
    }

  def _generate_examples(self, path):
    """Yields examples."""
    file_path = path['file_path']
    return rlds_base.generate_examples(file_path)



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Format-specific dataset builders for CoNLL-like formatted data.

It contains a ConllBuilderConfig and a ConllDatasetBuilder which are used to
initialize TFDS datasets based on CoNLL-like formatted data.
"""
from typing import Callable, List, Mapping, Optional, OrderedDict, Sequence, Union

from etils import epath
from tensorflow_datasets.core import dataset_builder
from tensorflow_datasets.core import dataset_info
from tensorflow_datasets.core import lazy_imports_lib
from tensorflow_datasets.core import split_builder as split_builder_lib
from tensorflow_datasets.core.features import feature as feature_lib
from tensorflow_datasets.core.features.features_dict import FeaturesDict


def get_conllu_example(
    sentence, example_id, features
) -> Mapping[str, Union[str, Sequence[str]]]:
  """Processes a conllu-annotated sentence into an example to be serialized.

  Args:
    sentence: the annotated sentence parsed with the conllu library.
    example_id: the example_id of the example, which will be used if the `idx`
      feature is present but not defined in the annotated sentence.
    features: the features defined in the output example.

  Returns:
    An example to be serialized.
  """
  example = {}

  for feature in features:
    # Use the idx parsed from the data, example_id if not available.
    if feature == "idx":
      idx = sentence.metadata.get("sent_id", example_id)
      example["idx"] = idx

    # CoNNL-U format stores tokens using the `form` tag.
    elif feature == "tokens":
      example["tokens"] = [token["form"] for token in sentence]

    # CoNNL-U format stores lemmas using the `lemma` tag.
    elif feature == "lemmas":
      example["lemmas"] = [token["lemma"] for token in sentence]

    elif feature == "text":
      if "text" in sentence.metadata:
        text = sentence.metadata["text"]
      else:
        text = " ".join(example["tokens"])
      example["text"] = text

    # All other features are Sequences whose feature name corresponds to
    # the respective tag name in the annotations generated with the
    # conllu library.
    else:
      # UPOS are stored as ClassLabels and are therefore not converted
      # into strings.
      # Future features might follow the same principle, therefore we
      # check for list membership.
      if feature in ["upos"]:
        example[feature] = [token[feature] for token in sentence]
      else:
        example[feature] = [str(token[feature]) for token in sentence]

  return example


def get_xtreme_pos_example(
    sentence, example_id, features
) -> Mapping[str, Union[str, Sequence[str]]]:
  """Processes an annotated sentence into an example for the xtreme_pos dataset.

  This function adds a further check ensuring that, at a given position in a
  sentence, both the token and the upos label are not empty.
  This is done for consistency with the xtreme implementation in other
  libraries, such as HuggingFace (rf. line 955):
  https://github.com/huggingface/datasets/blob/e6f1352fe19679de897f3d962e616936a17094f5/datasets/xtreme/xtreme.py

  Args:
    sentence: the annotated sentence parsed with the conllu library.
    example_id: the example_id of the example, which will be used if the `idx`
      feature is present but not defined in the annotated sentence.
    features: the features defined in the output example.

  Returns:
    An example to be serialized.
  """
  del example_id
  example = {feature: [] for feature in features}
  for token in sentence:
    if token["form"] != "_" and token["upos"] != "_":
      example["tokens"].append(token["form"])
      example["upos"].append(token["upos"])
  return example


# TODO(b/241346210): Should update ConllUBuilderConfig to @dataclasses.dataclass
class ConllUBuilderConfig(dataset_builder.BuilderConfig):
  """Base class for CoNLL-U formatted data configuration.

  Attributes:
    features: An OrderedDict specifying the features names and their type.
    language: The language of the data used to generate the ConllUBuilderConfig.
  """

  def __init__(
      self,
      *,
      features: OrderedDict[str, feature_lib.FeatureConnector],
      language: str,
      **kwargs,
  ):
    """Initializes the builder config for Conll-U formatted datasets.

    Args:
      features: An OrderedDict specifying the features names and their type.
      language: The language of the data used to generate the
        ConllUBuilderConfig.
      **kwargs: keyword arguments forwarded to super.
    """
    super(ConllUBuilderConfig, self).__init__(**kwargs)
    self.features = features
    self.language = language

  @property
  def features_dict(self) -> FeaturesDict:
    return FeaturesDict(self.features)


class ConllUDatasetBuilder(
    dataset_builder.GeneratorBasedBuilder, skip_registration=True
):
  """Base class for CoNLL-like formatted datasets.

  It provides functionalities to ease the processing of CoNLL-like datasets.
  Users can overwrite `_generate_examples` to customize the pipeline.
  """

  BUILDER_CONFIGS: Sequence[ConllUBuilderConfig] = []

  @property
  def builder_config(self) -> ConllUBuilderConfig:
    """`tfds.core.BuilderConfig` for this builder."""
    return self._builder_config  # pytype: disable=bad-return-type  # always-use-return-annotations

  def create_dataset_info(
      self,
      description: Optional[str] = None,
      supervised_keys: Optional[dataset_info.SupervisedKeysType] = None,
      homepage: Optional[str] = None,
      citation: Optional[str] = None,
  ) -> dataset_info.DatasetInfo:
    """Initializes `dataset_info.DatasetInfo` for Conll-U datasets.

    Args:
      description: [DEPRECATED] A short, markdown-formatted description of the
        dataset. Prefer placing description in `README.md` file.
      supervised_keys:  Specifies the input structure for supervised learning,
        if applicable for the dataset, used with "as_supervised". Typically this
        is a `(input_key, target_key)` tuple.
      homepage: The homepage of the dataset, if applicable for this dataset.
      citation: [DEPRECATED] The citation to use for this dataset, if applicable
        for this dataset. Prefer placing citations in `CITATIONS.bib` file.

    Returns:
      `dataset_info.DatasetInfo` for Conll-U datasets, populated with the values
      from the provided arguments.
    """
    return self.dataset_info_from_configs(
        description=description,
        features=self.builder_config.features_dict,
        supervised_keys=supervised_keys,
        homepage=homepage,
        citation=citation,
    )

  def _generate_examples(
      self,
      filepaths: Union[epath.PathLike, List[epath.PathLike]],
      process_example_fn: Callable[
          ..., Mapping[str, Union[str, Sequence[str]]]
      ] = get_conllu_example,
  ) -> split_builder_lib.SplitGenerator:
    """Processes CoNLL-U formatted datasets and generate examples.

    Args:
      filepaths: The filepaths of the input data. Could be a list of paths for
        multiple input files, or a single path.
      process_example_fn: The function used to process a conllu-annotated
        sentence into an example to be serialized. Defaults to
        get_conllu_example.

    Yields:
      Generated examples.
    """
    conllu = lazy_imports_lib.lazy_imports.conllu
    path = filepaths if isinstance(filepaths, list) else [filepaths]

    example_id = 0
    for filepath in path:
      with epath.Path(filepath).open() as data_file:
        annotated_sentences = list(conllu.parse_incr(data_file))
        for sentence in annotated_sentences:
          example = process_example_fn(
              sentence=sentence,
              example_id=example_id,
              features=self.builder_config.features,
          )
          yield example_id, example

          example_id += 1
